from .executable_finder import ExecutableFinder
from .executable_start_type import ExecutableStartType
from .live_packet_gather_parameters import LivePacketGatherParameters
from .provenance_data_item import ProvenanceDataItem

__all__ = ["ExecutableFinder", "ExecutableStartType",
           "LivePacketGatherParameters", "ProvenanceDataItem"]



from spinn_utilities.progress_bar import ProgressBar
from spinn_front_end_common.utilities.exceptions import ConfigurationException
import re
import os


ERROR_ENTRY = re.compile("\[ERROR\]\s+\((.*)\):\s+(.*)")
WARNING_ENTRY = re.compile("\[WARNING\]\s+\((.*)\):\s+(.*)")
ENTRY_FILE = 1
ENTRY_TEXT = 2


class ChipIOBufExtractor(object):
    """ Extract iobuf buffers from the machine, and separates lines based on\
        their prefix
    """

    __slots__ = []

    def __call__(
            self, transceiver, has_ran, core_subsets, provenance_file_path):
        if not has_ran:
            raise ConfigurationException(
                "The simulation needs to have tried to run before asking for"
                "iobuf. Please fix and try again")

        return self._run_for_core_subsets(
            core_subsets, transceiver, provenance_file_path)

    def _run_for_core_subsets(
            self, core_subsets, transceiver, provenance_file_path):
        progress = ProgressBar(
            len(core_subsets), "Extracting IOBUF from the machine")
        error_entries = list()
        warn_entries = list()

        # extract iobuf
        io_buffers = list(transceiver.get_iobuf(core_subsets))

        # write iobuf
        for iobuf in io_buffers:
            file_name = os.path.join(
                provenance_file_path,
                "iobuf_for_chip_{}_{}_processor_id_{}.txt".format(
                    iobuf.x, iobuf.y, iobuf.p))

            # set mode of the file based off if the file already exists
            mode = "w"
            if os.path.exists(file_name):
                mode = "a"

            # write iobuf to file.
            with open(file_name, mode) as writer:
                writer.write(iobuf.iobuf)

        # check iobuf for errors
        for io_buffer in progress.over(io_buffers):
            self._check_iobuf_for_error(io_buffer, error_entries, warn_entries)
        return error_entries, warn_entries

    def _check_iobuf_for_error(self, iobuf, error_entries, warn_entries):
        lines = iobuf.iobuf.split("\n")
        for line in lines:
            line = line.encode('ascii', 'ignore')
            self._add_value_if_match(
                ERROR_ENTRY, line, error_entries, iobuf.x, iobuf.y, iobuf.p)
            self._add_value_if_match(
                WARNING_ENTRY, line, warn_entries, iobuf.x, iobuf.y, iobuf.p)

    @staticmethod
    def _add_value_if_match(regex, line, entries, x, y, p):
        match = regex.match(line)
        if match:
            entries.append("{}, {}, {}: {} ({})".format(
                x, y, p, match.group(ENTRY_TEXT), match.group(ENTRY_FILE)))



import logging
import time

from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_front_end_common.utilities.utility_objs import ExecutableStartType

from spinnman.messages.scp.enums import Signal
from spinnman.model.enums import CPUState

logger = logging.getLogger(__name__)


class ApplicationRunner(object):
    """ Ensures all cores are initialised correctly, ran, and completed\
        successfully.
    """

    __slots__ = []

    def __call__(
            self, buffer_manager, wait_on_confirmation, send_stop_notification,
            send_start_notification, notification_interface,
            executable_targets, executable_start_type, app_id, txrx, runtime,
            time_scale_factor, loaded_reverse_iptags_token,
            loaded_iptags_token, loaded_routing_tables_token,
            loaded_binaries_token, loaded_application_data_token,
            no_sync_changes, time_threshold, run_until_complete=False):

        # check all tokens are valid
        if (not loaded_reverse_iptags_token or not loaded_iptags_token or
                not loaded_routing_tables_token or not loaded_binaries_token or
                not loaded_application_data_token):
            raise ConfigurationException(
                "Not all valid tokens have been given in the positive state")

        logger.info("*** Running simulation... *** ")

        # Get the expected state of the application, depending on the run type
        expected_states = None
        sync_signal = None
        if executable_start_type == ExecutableStartType.RUNNING:
            expected_states = [
                CPUState.RUNNING, CPUState.FINISHED, CPUState.PAUSED,
                CPUState.SYNC0, CPUState.SYNC1
            ]
        elif executable_start_type == ExecutableStartType.SYNC:
            sync_signal = Signal.SYNC0
            expected_states = [CPUState.SYNC0]
        elif (executable_start_type ==
                ExecutableStartType.USES_SIMULATION_INTERFACE):
            if no_sync_changes % 2 == 0:
                expected_states = [CPUState.SYNC0]
                sync_signal = Signal.SYNC0
            else:
                expected_states = [CPUState.SYNC1]
                sync_signal = Signal.SYNC1

            # when it falls out of the running, it'll be in a next sync state,
            # thus update needed
            no_sync_changes += 1

        if expected_states is None:
            raise ConfigurationException(
                "Unknown executable start type {}".format(
                    executable_start_type))

        # wait for all cores to be ready
        txrx.wait_for_cores_to_be_in_state(
            executable_targets.all_core_subsets, app_id, expected_states)

        # set the buffer manager into a resume state, so that if it had ran
        # before it'll work again
        buffer_manager.resume()

        # every thing is in sync0 so load the initial buffers
        buffer_manager.load_initial_buffers()

        # wait till external app is ready for us to start if required
        if notification_interface is not None and wait_on_confirmation:
            notification_interface.wait_for_confirmation()

        # set off the executables that are in sync state
        if sync_signal is not None:
            txrx.send_signal(app_id, sync_signal)
            txrx.wait_for_cores_to_be_in_state(
                executable_targets.all_core_subsets, app_id,
                [CPUState.RUNNING, CPUState.PAUSED, CPUState.FINISHED])

        # Send start notification
        if notification_interface is not None and send_start_notification:
            notification_interface.send_start_resume_notification()

        # Wait for the application to finish
        if runtime is None and not run_until_complete:
            logger.info("Application is set to run forever - exiting")
        else:
            timeout = None
            if not run_until_complete:
                time_to_wait = ((runtime * time_scale_factor) / 1000.0) + 0.1
                logger.info(
                    "Application started - waiting {} seconds for it to stop"
                    .format(time_to_wait))
                time.sleep(time_to_wait)
                timeout = time_threshold
            else:
                logger.info(
                    "Application started - waiting until finished")

            txrx.wait_for_cores_to_be_in_state(
                executable_targets.all_core_subsets, app_id,
                [CPUState.FINISHED, CPUState.PAUSED], timeout=timeout)

        if (notification_interface is not None and
                send_stop_notification and runtime is not None):
            notification_interface.send_stop_pause_notification()

        return True, no_sync_changes



from spinn_front_end_common.utilities.exceptions import ConfigurationException

from spinnman.model.enums import CPUState
from spinn_front_end_common.utilities.scp import UpdateRuntimeProcess


class ChipRuntimeUpdater(object):
    """ Updates the runtime of an application running on a spinnaker machine
    """

    __slots__ = []

    def __call__(
            self, txrx, app_id, executable_targets,
            no_machine_timesteps, loaded_binaries_token):
        if not loaded_binaries_token:
            raise ConfigurationException(
                "The binaries must be loaded before the run time updater is"
                " called")

        txrx.wait_for_cores_to_be_in_state(
            executable_targets.all_core_subsets, app_id, [CPUState.PAUSED])

        infinite_run = 0
        if no_machine_timesteps is None:
            infinite_run = 1
            no_machine_timesteps = 0

        # TODO: Expose the connection selector in SpiNNMan
        process = UpdateRuntimeProcess(txrx._scamp_connection_selector)
        process.update_runtime(
            no_machine_timesteps, infinite_run,
            executable_targets.all_core_subsets,
            executable_targets.total_processors)

        return True



import math
import logging
import sys
from threading import Thread

from spalloc import Job
from spalloc.states import JobState
from spinn_front_end_common.abstract_models \
    import AbstractMachineAllocationController

logger = logging.getLogger(__name__)


class _SpallocJobController(Thread, AbstractMachineAllocationController):

    __slots__ = [
        # thread flag to allow it to be killed when the main thread dies
        "daemon",

        # the spalloc job object
        "_job",

        # boolean flag for telling this thread when the system has ended
        "_exited"
    ]

    def __init__(self, job):
        Thread.__init__(self, name="SpallocJobController")
        self.daemon = True
        self._job = job
        self._exited = False

    def extend_allocation(self, new_total_run_time):
        # Does Nothing in this allocator - machines are held until exit
        pass

    def close(self):
        self._exited = True
        self._job.destroy()

    @property
    def power(self):
        return self._job.power

    def set_power(self, power):
        self._job.set_power(power)
        if power:
            self._job.wait_until_ready()

    def where_is_machine(self, chip_x, chip_y):
        return self._job.where_is_machine(chip_y=chip_y, chip_x=chip_x)

    def _wait_for_state_change(self, old_state):
        try:
            if self._job is not None:
                return self._job.wait_for_state_change(old_state)
        except TypeError:
            pass
        return old_state

    def run(self):
        state = self._job.state
        while state != JobState.destroyed and not self._exited:
            state = self._wait_for_state_change(state)

        self._job.close()

        if not self._exited:
            logger.error(
                "The allocated machine has been released before the end of"
                " the script - this script will now exit")
            sys.exit(1)


class SpallocAllocator(object):
    """ Request a machine from a SPALLOC server that will fit the given\
        number of chips
    """

    # Use a worst case calculation
    _N_CHIPS_PER_BOARD = 48.0
    _MACHINE_VERSION = 5

    def __call__(
            self, spalloc_server, spalloc_user, n_chips, spalloc_port=None,
            spalloc_machine=None):
        """

        :param spalloc_server: \
            The server from which the machine should be requested
        :param spalloc_port: The port of the SPALLOC server
        :param spalloc_user: The user to allocate the machine to
        :param n_chips: The number of chips required
        :param spalloc_port: The optional port number to speak to spalloc
        :param spalloc_machine: The optional spalloc machine to use
        """

        # Work out how many boards are needed
        n_boards = float(n_chips) / self._N_CHIPS_PER_BOARD

        # If the number of boards rounded up is less than 10% bigger than the
        # actual number of boards, add another board just in case
        if math.ceil(n_boards) - n_boards < 0.1:
            n_boards += 1
        n_boards = int(math.ceil(n_boards))

        spalloc_kw_args = {
            'hostname': spalloc_server,
            'owner': spalloc_user
        }
        if spalloc_port is not None:
            spalloc_kw_args['port'] = spalloc_port
        if spalloc_machine is not None:
            spalloc_kw_args['machine'] = spalloc_machine

        job, hostname = self._launch_job(n_boards, spalloc_kw_args)
        machine_allocation_controller = _SpallocJobController(job)
        machine_allocation_controller.start()

        return (
            hostname, self._MACHINE_VERSION, None, None, None, None, False,
            False, None, None, None, machine_allocation_controller
        )

    def _launch_job(self, n_boards, spalloc_kw_args):
        job = Job(n_boards, **spalloc_kw_args)
        try:
            job.wait_until_ready()
            # get param from jobs before starting, so that hanging doesn't
            # occur
            return job, job.hostname
        except:
            job.destroy()
            raise



class LivePacketGatherParameters(object):
    """ parameter holder for LPG's so that they can be instantiated at a\
     later date.
    """

    __slots__ = [
        '_port', '_hostname', "_tag", "_board_address", "_strip_sdp",
        "_use_prefix", "_key_prefix", "_prefix_type", "_message_type",
        "_right_shift", "_payload_as_time_stamps", "_use_payload_prefix",
        "_payload_prefix",  "_payload_right_shift",
        "_number_of_packets_sent_per_time_step", "_label", "_partition_id"
    ]

    def __init__(
            self, port, hostname, tag, board_address, strip_sdp, use_prefix,
            key_prefix, prefix_type, message_type, right_shift,
            payload_as_time_stamps, use_payload_prefix, payload_prefix,
            payload_right_shift, number_of_packets_sent_per_time_step,
            partition_id):

        self._port = port
        self._hostname = hostname
        self._tag = tag
        self._board_address = board_address
        self._strip_sdp = strip_sdp
        self._use_prefix = use_prefix
        self._key_prefix = key_prefix
        self._prefix_type = prefix_type
        self._message_type = message_type
        self._right_shift = right_shift
        self._payload_as_time_stamps = payload_as_time_stamps
        self._use_payload_prefix = use_payload_prefix
        self._payload_prefix = payload_prefix
        self._payload_right_shift = payload_right_shift
        self._number_of_packets_sent_per_time_step = \
            number_of_packets_sent_per_time_step
        self._partition_id = partition_id

    @property
    def port(self):
        return self._port

    @property
    def hostname(self):
        return self._hostname

    @property
    def tag(self):
        return self._tag

    @property
    def board_address(self):
        return self._board_address

    @property
    def strip_sdp(self):
        return self._strip_sdp

    @property
    def use_prefix(self):
        return self._use_prefix

    @property
    def key_prefix(self):
        return self._key_prefix

    @property
    def prefix_type(self):
        return self._prefix_type

    @property
    def message_type(self):
        return self._message_type

    @property
    def right_shift(self):
        return self._right_shift

    @property
    def payload_as_time_stamps(self):
        return self._payload_as_time_stamps

    @property
    def use_payload_prefix(self):
        return self._use_payload_prefix

    @property
    def payload_prefix(self):
        return self._payload_prefix

    @property
    def payload_right_shift(self):
        return self._payload_right_shift

    @property
    def number_of_packets_sent_per_time_step(self):
        return self._number_of_packets_sent_per_time_step

    @property
    def partition_id(self):
        return self._partition_id

    def __eq__(self, other):
        if (self._port == other.port and self._hostname == other.hostname and
                self._tag == other.tag and
                self._board_address == other.board_address and
                self._strip_sdp == other.strip_sdp and
                self._use_prefix == other.use_prefix and
                self._key_prefix == other.key_prefix and
                self._prefix_type == other.prefix_type and
                self._message_type == other.message_type and
                self._right_shift == other.right_shift and
                self._payload_as_time_stamps ==
                other.payload_as_time_stamps and
                self._use_payload_prefix == other.use_payload_prefix and
                self._payload_prefix == other.payload_prefix and
                self._payload_right_shift == other.payload_right_shift and
                self._number_of_packets_sent_per_time_step ==
                other.number_of_packets_sent_per_time_step and
                self._partition_id == other.partition_id):
            return True
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        data = (
            self._port, self._tag, self._board_address, self._strip_sdp,
            self._use_prefix, self._key_prefix, self._prefix_type,
            self._message_type, self._right_shift,
            self._payload_as_time_stamps, self._use_payload_prefix,
            self._payload_prefix, self._payload_right_shift,
            self._number_of_packets_sent_per_time_step, self._partition_id)
        return hash(data)



from spinn_utilities.progress_bar import ProgressBar

from spinn_front_end_common.utilities.exceptions import SpinnFrontEndException
from spinn_front_end_common.mapping_algorithms \
    import on_chip_router_table_compression
from spinn_front_end_common.interface.interface_functions \
    import ChipIOBufExtractor

from spinnman.model.enums import CPUState
from spinnman.model import ExecutableTargets

from spinn_machine import CoreSubsets, Router

import logging
import os
import struct

logger = logging.getLogger(__name__)

# The SDRAM Tag used by the application - note this is fixed in the APLX
_SDRAM_TAG = 1


class MundyOnChipRouterCompression(object):
    """ Compressor that uses a on chip router compressor
    """

    SIZE_OF_A_SDRAM_ENTRY = 4 * 4
    SURPLUS_DATA_ENTRIES = 3 * 4
    TIME_EXPECTED_TO_RUN = 1000
    OVER_RUN_THRESHOLD_BEFORE_ERROR = 1000

    def __call__(
            self, routing_tables, transceiver, machine, app_id,
            provenance_file_path, compress_only_when_needed=True,
            compress_as_much_as_possible=False):
        """

        :param routing_tables: the memory routing tables to be compressed
        :param transceiver: the spinnman interface
        :param machine: the spinnaker machine representation
        :param app_id: the app-id used by the main application
        :param provenance_file_path: the path to where to write the data
        :return: flag stating routing compression and loading has been done
        """

        # build progress bar
        progress = ProgressBar(
            len(routing_tables.routing_tables) + 2,
            "Running routing table compression on chip")
        compressor_app_id = transceiver.app_id_tracker.get_new_id()

        # figure size of sdram needed for each chip for storing the routing
        # table
        for routing_table in routing_tables:
            self._load_routing_table(
                routing_table, transceiver, app_id, compressor_app_id,
                compress_only_when_needed, compress_as_much_as_possible)
            # update progress bar
            progress.update()

        # load the router compressor executable
        executable_targets = self._load_executables(
            routing_tables, compressor_app_id, transceiver, machine)

        # update progress bar
        progress.update()

        # Wait for the executable to finish
        succeeded = False
        try:
            transceiver.wait_for_cores_to_be_in_state(
                executable_targets.all_core_subsets, compressor_app_id,
                [CPUState.FINISHED])
            succeeded = True
        finally:
            # get the debug data
            if not succeeded:
                self._handle_failure(
                    executable_targets, transceiver, provenance_file_path,
                    compressor_app_id)

        # Check if any cores have not completed successfully
        self._check_for_success(
            executable_targets, transceiver,
            provenance_file_path, compressor_app_id)

        # update progress bar
        progress.update()

        # stop anything that's associated with the compressor binary
        transceiver.stop_application(compressor_app_id)
        transceiver.app_id_tracker.free_id(compressor_app_id)

        # update the progress bar
        progress.end()

        # return loaded routing tables flag
        return True

    def _load_routing_table(
            self, table, txrx, app_id, compressor_app_id,
            compress_only_when_needed, compress_as_much_as_possible):
        data = self._build_data(
            table, app_id, compress_only_when_needed,
            compress_as_much_as_possible)

        # go to spinnman and ask for a memory region of that size per chip.
        base_address = txrx.malloc_sdram(
            table.x, table.y, len(data), compressor_app_id, _SDRAM_TAG)

        # write sdram requirements per chip
        txrx.write_memory(table.x, table.y, base_address, data)

    def _check_for_success(
            self, executable_targets, transceiver, provenance_file_path,
            compressor_app_id):
        """ goes through the cores checking for cores that have failed to\
            compress the routing tables to the level where they fit into the\
            router
        """

        for core_subset in executable_targets.all_core_subsets:
            x = core_subset.x
            y = core_subset.y
            for p in core_subset.processor_ids:

                # Read the result from USER0 register
                user_0_address = \
                    transceiver.get_user_0_register_address_from_core(x, y, p)

                result = struct.unpack(
                    "<I", str(transceiver.read_memory(x, y, user_0_address, 4))
                )[0]

                # The result is 0 if success, otherwise failure
                if result != 0:
                    self._handle_failure(
                        executable_targets, transceiver, provenance_file_path,
                        compressor_app_id)

                    raise SpinnFrontEndException(
                        "The router compressor on {}, {} failed to complete"
                        .format(x, y))

    def _handle_failure(
            self, executable_targets, transceiver, provenance_file_path,
            compressor_app_id):
        """

        :param executable_targets:
        :param transceiver:
        :param provenance_file_path:
        :rtype: None
        """
        logger.info("Router compressor has failed")
        iobuf_extractor = ChipIOBufExtractor()
        io_buffers, io_errors, io_warnings = iobuf_extractor(
            transceiver, True, executable_targets.all_core_subsets)
        self._write_iobuf(io_buffers, provenance_file_path)
        for warning in io_warnings:
            logger.warn(warning)
        for error in io_errors:
            logger.error(error)
        transceiver.stop_application(compressor_app_id)
        transceiver.app_id_tracker.free_id(compressor_app_id)

    @staticmethod
    def _write_iobuf(io_buffers, provenance_file_path):
        """ writes the iobuf to files

        :param io_buffers: the iobuf for the cores
        :param provenance_file_path:\
            the file path where the iobuf are to be stored
        :rtype: None
        """
        for iobuf in io_buffers:
            file_name = os.path.join(
                provenance_file_path,
                "{}_{}_{}_compressor.txt".format(iobuf.x, iobuf.y, iobuf.p))
            count = 2
            while os.path.exists(file_name):
                file_name = os.path.join(
                    provenance_file_path,
                    "{}_{}_{}_compressor-{}.txt".format(
                        iobuf.x, iobuf.y, iobuf.p, count))
                count += 1
            with open(file_name, "w") as writer:
                writer.write(iobuf.iobuf)

    def _load_executables(
            self, routing_tables, compressor_app_id, transceiver, machine):
        """ loads the router compressor onto the chips.

        :param routing_tables: the router tables needed to be compressed
        :param compressor_app_id: the app id of the compressor compressor
        :param transceiver: the spinnman interface
        :param machine: the spinnaker machine representation
        :return:\
            the executable targets that represent all cores/chips which have\
            active routing tables
        """

        # build core subsets
        core_subsets = CoreSubsets()
        for routing_table in routing_tables:

            # get the first none monitor core
            chip = machine.get_chip_at(routing_table.x, routing_table.y)
            processor = chip.get_first_none_monitor_processor()

            # add to the core subsets
            core_subsets.add_processor(
                routing_table.x, routing_table.y, processor.processor_id)

        # build binary path
        binary_path = os.path.join(
            os.path.dirname(on_chip_router_table_compression.__file__),
            "rt_minimise.aplx")

        # build executable targets
        executable_targets = ExecutableTargets()
        executable_targets.add_subsets(binary_path, core_subsets)

        transceiver.execute_application(executable_targets, compressor_app_id)
        return executable_targets

    def _build_data(
            self, routing_table, app_id, compress_only_when_needed,
            compress_as_much_as_possible):
        """ convert the router table into the data needed by the router\
            compressor c code.

        :param routing_table: the pacman router table instance
        :param app_id: the app-id to load the entries in by
        :param compress_only_when_needed:\
            If True, the compressor will only compress if the table doesn't\
            fit in the current router space, otherwise it will just load\
            the table
        :param compress_as_much_as_possible:\
            If False, the compressor will only reduce the table until it fits\
            in the router space, otherwise it will try to reduce until it\
            until it can't reduce it any more
        :return: The byte array of data
        """

        # write header data of the app id to load the data, if to store
        # results in sdram and the router table entries

        data = b''
        data += struct.pack("<I", app_id)
        data += struct.pack("<I", int(compress_only_when_needed))
        data += struct.pack("<I", int(compress_as_much_as_possible))

        # Write the size of the table
        data += struct.pack("<I", routing_table.number_of_entries)

        for entry in routing_table.multicast_routing_entries:
            data += struct.pack("<I", entry.routing_entry_key)
            data += struct.pack("<I", entry.mask)
            data += struct.pack(
                "<I",
                Router.convert_routing_table_entry_to_spinnaker_route(entry))
            data += struct.pack("<I", self._make_source_hack(entry))
        return bytearray(data)

    @staticmethod
    def _make_source_hack(entry):
        """ Hack to support the source requirement for the router compressor\
            on chip

        :param entry: the multicast router table entry.
        :return: return the source value
        """
        if entry.defaultable:
            return list(entry.link_ids)[0] + 3 % 6
        elif len(entry.link_ids) > 0:
            return list(entry.link_ids)[0]
        else:
            return 0



from spinn_utilities.progress_bar import ProgressBar
from spinnman.constants import MAX_TAG_ID


class TagsLoader(object):
    """ Loads tags onto the machine
    """

    __slots__ = []

    def __call__(
            self, transceiver, tags=None, iptags=None, reverse_iptags=None):
        """
        :param tags: the tags object which contains ip and reverse ip tags.
                    could be none if these are being given in separate lists
        :param iptags: a list of iptags, given when tags is none
        :param reverse_iptags: a list of reverse iptags when tags is none.
        :param transceiver: the transceiver object
        """
        # clear all the tags from the Ethernet connection, as nothing should
        # be allowed to use it (no two apps should use the same Ethernet
        # connection at the same time
        progress = ProgressBar(MAX_TAG_ID, "Clearing tags")

        for tag_id in progress.over(range(MAX_TAG_ID)):
            transceiver.clear_ip_tag(tag_id)

        progress = None
        if tags is not None:
            progress = ProgressBar(
                len(list(tags.ip_tags)) + len(list(tags.reverse_ip_tags)),
                "Loading Tags")
            self.load_iptags(tags.ip_tags, transceiver, progress)
            self.load_reverse_iptags(
                tags.reverse_ip_tags, transceiver, progress)
        else:
            progress = ProgressBar(
                len(iptags) + len(reverse_iptags),
                "Loading Tags")
            self.load_iptags(iptags, transceiver, progress)
            self.load_reverse_iptags(reverse_iptags, transceiver, progress)
        progress.end()

        return True, True

    @staticmethod
    def load_iptags(iptags, transceiver, progress_bar):
        """ Loads all the iptags individually.

        :param iptags: the iptags to be loaded.
        :param transceiver: the transceiver object
        :rtype: None
        """
        for ip_tag in iptags:
            transceiver.set_ip_tag(ip_tag)
            progress_bar.update()

    @staticmethod
    def load_reverse_iptags(reverse_ip_tags, transceiver, progress_bar):
        """ Loads all the reverse iptags individually.

        :param reverse_ip_tags: the reverse iptags to be loaded
        :param transceiver: the transceiver object
        :rtype: None
        """
        for reverse_ip_tag in reverse_ip_tags:
            transceiver.set_reverse_ip_tag(reverse_ip_tag)
            progress_bar.update()



from spinnman.messages.sdp import SDPHeader, SDPFlag
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.impl import CheckOKResponse
from spinn_front_end_common.utilities.constants \
    import SDP_RUNNING_MESSAGE_CODES


class SCPUpdateRuntimeRequest(AbstractSCPRequest):

    def __init__(
            self, x, y, p, run_time, infinite_run, destination_port,
            expect_response=True):
        sdp_flags = SDPFlag.REPLY_NOT_EXPECTED
        arg3 = 0
        if expect_response:
            sdp_flags = SDPFlag.REPLY_EXPECTED
            arg3 = 1

        AbstractSCPRequest.__init__(
            self,
            SDPHeader(
                flags=sdp_flags, destination_port=destination_port,
                destination_cpu=p, destination_chip_x=x, destination_chip_y=y),
            SCPRequestHeader(
                command=SDP_RUNNING_MESSAGE_CODES.SDP_NEW_RUNTIME_ID_CODE),
            argument_1=run_time, argument_2=infinite_run, argument_3=arg3)

    def get_scp_response(self):
        return CheckOKResponse(
            "update runtime",
            SDP_RUNNING_MESSAGE_CODES.SDP_NEW_RUNTIME_ID_CODE.value)



from spinn_utilities.progress_bar import ProgressBar

from spinn_front_end_common.interface.buffer_management import BufferManager
from spinn_front_end_common.interface.buffer_management.buffer_models \
    import AbstractSendsBuffersFromHost, AbstractReceiveBuffersToHost


class BufferManagerCreator(object):
    __slots__ = []

    def __call__(self, placements, tags, txrx, store_data_in_file):
        progress = ProgressBar(placements.placements, "Initialising buffers")

        # Create the buffer manager
        buffer_manager = BufferManager(
            placements, tags, txrx, store_data_in_file)

        for placement in progress.over(placements.placements):
            if isinstance(placement.vertex, AbstractSendsBuffersFromHost):
                if placement.vertex.buffering_input():
                    buffer_manager.add_sender_vertex(placement.vertex)

            if isinstance(placement.vertex, AbstractReceiveBuffersToHost):
                buffer_manager.add_receiving_vertex(placement.vertex)

        return buffer_manager




_failed_state = None
_simulator = None


def get_simulator():
    global _simulator, _failed_state
    if _simulator is None:
        if _failed_state is None:
            raise ValueError("You must import one of the simulator classes "
                             "before calling get_simulator")
        return _failed_state
    return _simulator


def get_not_running_simulator():
    global _simulator, _failed_state
    if _simulator is None:
        if _failed_state is None:
            raise ValueError("You must import one of the simulator classes "
                             "before calling get_simulator")
        return _failed_state
    _simulator.verify_not_running()
    return _simulator


def set_simulator(new_simulator):
    global _simulator, _failed_state
    if _failed_state is None:
        raise ValueError("Unexpected call to set_simulator before "
                         "set_failed_state")
    _simulator = new_simulator


def unset_simulator():
    global _simulator
    _simulator = None


def has_simulator():
    global _simulator
    return _simulator is not None


def set_failed_state(new_failed_state):
    global _failed_state
    if _failed_state is None:
        _failed_state = new_failed_state
    else:
        if type(new_failed_state) != type(_failed_state):
            raise ValueError("You may only setup/init one type of simulator")



from spinn_front_end_common.utilities.scp import ClearIOBUFProcess

from spinn_front_end_common.utilities.exceptions import ConfigurationException


class ChipIOBufClearer(object):
    """ Updates the runtime of an application running on a spinnaker machine
    """

    __slots__ = []

    def __call__(self, transceiver, executable_targets, ran_token):
        if not ran_token:
            raise ConfigurationException(
                "The simulation has to have ran before running this system")

        process = ClearIOBUFProcess(transceiver._scamp_connection_selector)
        process.clear_iobuf(executable_targets.all_core_subsets,
                            len(executable_targets.all_core_subsets))
        return True



# spinn_utilites imports
from spinn_utilities.progress_bar import ProgressBar

# spinnman imports
from spinnman.constants import UDP_MESSAGE_MAX_SIZE
from spinnman.connections.udp_packet_connections import EIEIOConnection
from spinnman.messages.eieio.command_messages \
    import EIEIOCommandMessage, StopRequests, SpinnakerRequestReadData
from spinnman.messages.eieio.command_messages \
    import HostDataRead, SpinnakerRequestBuffers, PaddingRequest
from spinnman.messages.eieio.command_messages \
    import HostSendSequencedData, EventStopRequest
from spinnman.utilities import utility_functions
from spinnman.messages.sdp import SDPHeader, SDPMessage, SDPFlag
from spinnman.exceptions import SpinnmanInvalidPacketException
from spinnman.messages.eieio import EIEIOType, create_eieio_command
from spinnman.messages.eieio.data_messages import EIEIODataMessage

# front end common imports
from spinn_front_end_common.utilities import helpful_functions as funs
from spinn_front_end_common.utilities import exceptions
from spinn_front_end_common.interface.buffer_management.storage_objects \
    import BuffersSentDeque, BufferedReceivingData, ChannelBufferState
from spinn_front_end_common.utilities.constants \
    import SDP_PORTS, BUFFERING_OPERATIONS
from .recording_utilities import TRAFFIC_IDENTIFIER, \
    get_last_sequence_number, get_region_pointer

# general imports
import threading
import logging


logger = logging.getLogger(__name__)

# The minimum size of any message - this is the headers plus one entry
_MIN_MESSAGE_SIZE = EIEIODataMessage.min_packet_length(
    eieio_type=EIEIOType.KEY_32_BIT, is_timestamp=True)

# The number of bytes in each key to be sent
_N_BYTES_PER_KEY = EIEIOType.KEY_32_BIT.key_bytes  # @UndefinedVariable


class BufferManager(object):
    """ Manager of send buffers
    """

    __slots__ = [
        # placements object
        "_placements",

        # list of tags
        "_tags",

        # SpiNNMan instance
        "_transceiver",

        # Set of (ip_address, port) that are being listened to for the tags
        "_seen_tags",

        # Set of vertices with buffers to be sent
        "_sender_vertices",

        # Dictionary of sender vertex -> buffers sent
        "_sent_messages",

        # storage area for received data from cores
        "_received_data",

        # Lock to avoid multiple messages being processed at the same time
        "_thread_lock_buffer_out",

        # Lock to avoid multiple messages being processed at the same time
        "_thread_lock_buffer_in",

        # bool flag
        "_finished",

        # listener port
        "_listener_port",

        # Store to file flag
        "_store_to_file"
    ]

    def __init__(self, placements, tags, transceiver, store_to_file=False):
        """

        :param placements: The placements of the vertices
        :type placements:\
                    :py:class:`pacman.model.placements.Placements`
        :param tags: The tags assigned to the vertices
        :type tags: :py:class:`pacman.model.tags.Tags`
        :param transceiver: The transceiver to use for sending and receiving\
                    information
        :type transceiver: :py:class:`spinnman.transceiver.Transceiver`
        :param store_to_file: True if the data should be temporarily stored\
                    in a file instead of in RAM (default uses RAM)
        :type store_to_file: bool
        """

        self._placements = placements
        self._tags = tags
        self._transceiver = transceiver

        # Set of (ip_address, port) that are being listened to for the tags
        self._seen_tags = set()

        # Set of vertices with buffers to be sent
        self._sender_vertices = set()

        # Dictionary of sender vertex -> buffers sent
        self._sent_messages = dict()

        # storage area for received data from cores
        self._received_data = BufferedReceivingData(store_to_file)
        self._store_to_file = store_to_file

        # Lock to avoid multiple messages being processed at the same time
        self._thread_lock_buffer_out = threading.Lock()
        self._thread_lock_buffer_in = threading.Lock()

        self._finished = False

        self._listener_port = None

    def receive_buffer_command_message(self, packet):
        """ Handle an EIEIO command message for the buffers

        :param packet: The eieio message received
        :type packet:\
                    :py:class:`spinnman.messages.eieio.command_messages.eieio_command_message.EIEIOCommandMessage`
        """
        try:
            if not self._finished:
                if isinstance(packet, SpinnakerRequestBuffers):
                    with self._thread_lock_buffer_in:
                        vertex = self._placements.get_vertex_on_processor(
                            packet.x, packet.y, packet.p)

                        if vertex in self._sender_vertices:

                            # logger.debug(
                            #     "received send request with sequence: {1:d},"
                            #     " space available: {0:d}".format(
                            #         packet.space_available,
                            #         packet.sequence_no))

                            # noinspection PyBroadException
                            try:
                                self._send_messages(
                                    packet.space_available, vertex,
                                    packet.region_id, packet.sequence_no)
                            except Exception:
                                logger.warn("problem when sending messages",
                                            exc_info=True)
                elif isinstance(packet, SpinnakerRequestReadData):
                    with self._thread_lock_buffer_out:

                        # logger.debug(
                        #     "received {} read request(s) with sequence: {},"
                        #     " from chip ({},{}, core {}".format(
                        #         packet.n_requests, packet.sequence_no,
                        #         packet.x, packet.y, packet.p))
                        try:
                            self._retrieve_and_store_data(packet)
                        except Exception:
                            logger.warn("problem when handling data",
                                        exc_info=True)
                elif isinstance(packet, EIEIOCommandMessage):
                    raise SpinnmanInvalidPacketException(
                        str(packet.__class__),
                        "The command packet is invalid for buffer management: "
                        "command id {0:d}".format(packet.eieio_header.command))
                else:
                    raise SpinnmanInvalidPacketException(
                        packet.__class__,
                        "The command packet is invalid for buffer management")
        except Exception:
            logger.warn("problem when processing received packet",
                        exc_info=True)

    def _create_connection(self, tag):
        connection = self._transceiver.register_udp_listener(
            self.receive_buffer_command_message, EIEIOConnection,
            local_port=tag.port, local_host=tag.ip_address)
        self._seen_tags.add((tag.ip_address, connection.local_port))
        utility_functions.send_port_trigger_message(
            connection, tag.board_address)
        logger.info(
            "Listening for packets using tag {} on {}:{}".format(
                tag.tag, connection.local_ip_address,
                connection.local_port))
        return connection

    def _add_buffer_listeners(self, vertex):
        """ Add listeners for buffered data for the given vertex
        """

        # Find a tag for receiving buffer data
        tags = self._tags.get_ip_tags_for_vertex(vertex)

        if tags is not None:
            # locate tag associated with the buffer manager traffic
            for tag in tags:
                if tag.traffic_identifier == TRAFFIC_IDENTIFIER:
                    # If the tag port is not assigned create a connection\
                    # and assign the port.  Note that this *should* \
                    # update the port number in any tags being shared
                    if tag.port is None:
                        # If connection already setup, ensure subsequent
                        # boards use same listener port in their tag
                        if self._listener_port is None:
                            connection = self._create_connection(tag)
                            tag.port = connection.local_port
                            self._listener_port = connection.local_port
                        else:
                            tag.port = self._listener_port

                    # In case we have tags with different specified ports,\
                    # also allow the tag to be created here
                    elif (tag.ip_address, tag.port) not in self._seen_tags:
                        self._create_connection(tag)

    def add_receiving_vertex(self, vertex):
        """ Add a vertex into the managed list for vertices\
            which require buffers to be received from them during runtime
        """
        self._add_buffer_listeners(vertex)

    def add_sender_vertex(self, vertex):
        """ Add a vertex into the managed list for vertices
            which require buffers to be sent to them during runtime

        :param vertex: the vertex to be managed
        :type vertex:\
                    :py:class:`spinnaker.pyNN.models.abstract_models.buffer_models.AbstractSendsBuffersFromHost`
        """
        self._sender_vertices.add(vertex)
        self._add_buffer_listeners(vertex)

    def load_initial_buffers(self):
        """ Load the initial buffers for the senders using mem writes
        """
        total_data = 0
        for vertex in self._sender_vertices:
            for region in vertex.get_regions():
                total_data += vertex.get_region_buffer_size(region)

        progress = ProgressBar(
            total_data, "Loading buffers ({} bytes)".format(total_data))
        for vertex in self._sender_vertices:
            for region in vertex.get_regions():
                self._send_initial_messages(vertex, region, progress)
        progress.end()

    def reset(self):
        """ Resets the buffered regions to start transmitting from the\
            beginning of its expected regions and clears the buffered out data\
            files
        """
        # reset buffered out
        self._received_data = BufferedReceivingData(self._store_to_file)

        # rewind buffered in
        for vertex in self._sender_vertices:
            for region in vertex.get_regions():
                vertex.rewind(region)

    def resume(self):
        """ Resets any data structures needed before starting running again
        """

        # update the received data items
        self._received_data.resume()

    def clear_recorded_data(self, x, y, p, recording_region_id):
        """ Removes the recorded data stored in memory.

        :param x: placement x coord
        :param y: placement y coord
        :param p: placement p coord
        :param recording_region_id: the recording region id

        """
        self._received_data.clear(x, y, p, recording_region_id)

    def _generate_end_buffering_state_from_machine(
            self, placement, state_region_base_address):

        # retrieve channel state memory area
        channel_state_data = str(self._transceiver.read_memory(
            placement.x, placement.y, state_region_base_address,
            ChannelBufferState.size_of_channel_state()))
        return ChannelBufferState.create_from_bytearray(channel_state_data)

    def _create_message_to_send(self, size, vertex, region):
        """ Creates a single message to send with the given boundaries.

        :param size: The number of bytes available for the whole packet
        :type size: int
        :param vertex: The vertex to get the keys from
        :type vertex:\
                    :py:class:`spynnaker.pyNN.models.abstract_models.buffer_models.AbstractSendsBuffersFromHost`
        :param region: The region of the vertex to get keys from
        :type region: int
        :return: A new message, or None if no keys can be added
        :rtype: None or\
                    :py:class:`spinnman.messages.eieio.data_messages.EIEIODataMessage`
        """

        # If there are no more messages to send, return None
        if not vertex.is_next_timestamp(region):
            return None

        # Create a new message
        next_timestamp = vertex.get_next_timestamp(region)
        message = EIEIODataMessage.create(
            EIEIOType.KEY_32_BIT, timestamp=next_timestamp)

        # If there is no room for the message, return None
        if message.size + _N_BYTES_PER_KEY > size:
            return None

        # Add keys up to the limit
        bytes_to_go = size - message.size
        while (bytes_to_go >= _N_BYTES_PER_KEY and
                vertex.is_next_key(region, next_timestamp)):

            key = vertex.get_next_key(region)
            message.add_key(key)
            bytes_to_go -= _N_BYTES_PER_KEY

        return message

    def _send_initial_messages(self, vertex, region, progress):
        """ Send the initial set of messages

        :param vertex: The vertex to get the keys from
        :type vertex:\
                    :py:class:`spynnaker.pyNN.models.abstract_models.buffer_models.AbstractSendsBuffersFromHost`
        :param region: The region to get the keys from
        :type region: int
        :return: A list of messages
        :rtype: list of\
                    :py:class:`spinnman.messages.eieio.data_messages.EIEIODataMessage`
        """

        # Get the vertex load details
        # region_base_address = self._locate_region_address(region, vertex)
        region_base_address = funs.locate_memory_region_for_placement(
            self._placements.get_placement_of_vertex(vertex), region,
            self._transceiver)
        placement = self._placements.get_placement_of_vertex(vertex)

        # Add packets until out of space
        sent_message = False
        bytes_to_go = vertex.get_region_buffer_size(region)
        if bytes_to_go % 2 != 0:
            raise exceptions.SpinnFrontEndException(
                "The buffer region of {} must be divisible by 2".format(
                    vertex))
        all_data = ""
        if vertex.is_empty(region):
            sent_message = True
        else:
            min_size_of_packet = _MIN_MESSAGE_SIZE
            while (vertex.is_next_timestamp(region) and
                    bytes_to_go > min_size_of_packet):
                space_available = min(bytes_to_go, 280)
                next_message = self._create_message_to_send(
                    space_available, vertex, region)
                if next_message is None:
                    break

                # Write the message to the memory
                data = next_message.bytestring
                all_data += data
                sent_message = True

                # Update the positions
                bytes_to_go -= len(data)
                progress.update(len(data))

        if not sent_message:
            raise exceptions.BufferableRegionTooSmall(
                "The buffer size {} is too small for any data to be added for"
                " region {} of vertex {}".format(bytes_to_go, region, vertex))

        # If there are no more messages and there is space, add a stop request
        if (not vertex.is_next_timestamp(region) and
                bytes_to_go >= EventStopRequest.get_min_packet_length()):
            data = EventStopRequest().bytestring
            # logger.debug(
            #    "Writing stop message of {} bytes to {} on {}, {}, {}".format(
            #         len(data), hex(region_base_address),
            #         placement.x, placement.y, placement.p))
            all_data += data
            bytes_to_go -= len(data)
            progress.update(len(data))
            self._sent_messages[vertex] = BuffersSentDeque(
                region, sent_stop_message=True)

        # If there is any space left, add padding
        if bytes_to_go > 0:
            padding_packet = PaddingRequest()
            n_packets = bytes_to_go / padding_packet.get_min_packet_length()
            data = padding_packet.bytestring
            data *= n_packets
            all_data += data

        # Do the writing all at once for efficiency
        self._transceiver.write_memory(
            placement.x, placement.y, region_base_address, all_data)

    def _send_messages(self, size, vertex, region, sequence_no):
        """ Send a set of messages
        """

        # Get the sent messages for the vertex
        if vertex not in self._sent_messages:
            self._sent_messages[vertex] = BuffersSentDeque(region)
        sent_messages = self._sent_messages[vertex]

        # If the sequence number is outside the window, return no messages
        if not sent_messages.update_last_received_sequence_number(sequence_no):
            return list()

        # Remote the existing packets from the size available
        bytes_to_go = size
        for message in sent_messages.messages:
            if isinstance(message.eieio_data_message, EIEIODataMessage):
                bytes_to_go -= message.eieio_data_message.size
            else:
                bytes_to_go -= (message.eieio_data_message
                                .get_min_packet_length())

        # Add messages up to the limits
        while (vertex.is_next_timestamp(region) and
                not sent_messages.is_full and bytes_to_go > 0):

            space_available = min(
                bytes_to_go,
                UDP_MESSAGE_MAX_SIZE -
                HostSendSequencedData.get_min_packet_length())
            # logger.debug(
            #     "Bytes to go {}, space available {}".format(
            #         bytes_to_go, space_available))
            next_message = self._create_message_to_send(
                space_available, vertex, region)
            if next_message is None:
                break
            sent_messages.add_message_to_send(next_message)
            bytes_to_go -= next_message.size
            # logger.debug("Adding additional buffer of {} bytes".format(
            #     next_message.size))

        # If the vertex is empty, send the stop messages if there is space
        if (not sent_messages.is_full and
                not vertex.is_next_timestamp(region) and
                bytes_to_go >= EventStopRequest.get_min_packet_length()):
            sent_messages.send_stop_message()

        # If there are no more messages, turn off requests for more messages
        if not vertex.is_next_timestamp(region) and sent_messages.is_empty():
            # logger.debug("Sending stop")
            self._send_request(vertex, StopRequests())

        # Send the messages
        for message in sent_messages.messages:
            # logger.debug("Sending message with sequence {}".format(
            #     message.sequence_no))
            self._send_request(vertex, message)

    def _send_request(self, vertex, message):
        """ Sends a request

        :param vertex: The vertex to send to
        :param message: The message to send
        """

        placement = self._placements.get_placement_of_vertex(vertex)
        sdp_header = SDPHeader(
            destination_chip_x=placement.x, destination_chip_y=placement.y,
            destination_cpu=placement.p, flags=SDPFlag.REPLY_NOT_EXPECTED,
            destination_port=SDP_PORTS.INPUT_BUFFERING_SDP_PORT.value)
        sdp_message = SDPMessage(sdp_header, message.bytestring)
        self._transceiver.send_sdp_message(sdp_message)

    def stop(self):
        """ Indicates that the simulation has finished, so no further\
            outstanding requests need to be processed
        """
        with self._thread_lock_buffer_in:
            with self._thread_lock_buffer_out:
                self._finished = True

    def get_data_for_vertex(self, placement, recording_region_id):
        """ Get a pointer to the data container for all the data retrieved\
            during the simulation from a specific region area of a core

        :param placement: the placement to get the data from
        :type placement: pacman.model.placements.Placement
        :param recording_region_id: desired recording data region
        :type recording_region_id: int
        :return: pointer to a class which inherits from\
                AbstractBufferedDataStorage
        :rtype:\
                :py:class:`spinn_front_end_common.interface.buffer_management.buffer_models.AbstractBufferedDataStorage`
        """

        recording_data_address = \
            placement.vertex.get_recording_region_base_address(
                self._transceiver, placement)

        # Ensure the last sequence number sent has been retrieved
        if not self._received_data.is_end_buffering_sequence_number_stored(
                placement.x, placement.y, placement.p):
            self._received_data.store_end_buffering_sequence_number(
                placement.x, placement.y, placement.p,
                get_last_sequence_number(
                    placement, self._transceiver, recording_data_address))

        # Read the data if not already received
        if not self._received_data.is_data_from_region_flushed(
                placement.x, placement.y, placement.p,
                recording_region_id):

            # Read the end state of the recording for this region
            if not self._received_data.is_end_buffering_state_recovered(
                    placement.x, placement.y, placement.p,
                    recording_region_id):

                end_state_address = get_region_pointer(
                    placement, self._transceiver, recording_data_address,
                    recording_region_id)
                end_state = self._generate_end_buffering_state_from_machine(
                    placement, end_state_address)
                self._received_data.store_end_buffering_state(
                    placement.x, placement.y, placement.p, recording_region_id,
                    end_state)
            else:
                end_state = self._received_data.\
                    get_end_buffering_state(
                        placement.x, placement.y, placement.p,
                        recording_region_id)

            start_ptr = end_state.start_address
            write_ptr = end_state.current_write
            end_ptr = end_state.end_address
            read_ptr = end_state.current_read

            # current read needs to be adjusted in case the last portion of the
            # memory has already been read, but the HostDataRead packet has not
            # been processed by the chip before simulation finished
            # This situation is identified by the sequence number of the last
            # packet sent to this core and the core internal state of the
            # output buffering finite state machine
            seq_no_last_ack_packet = \
                self._received_data.last_sequence_no_for_core(
                    placement.x, placement.y, placement.p)

            # get the last sequence number
            last_sequence_number = \
                self._received_data.get_end_buffering_sequence_number(
                    placement.x, placement.y, placement.p)

            if last_sequence_number == seq_no_last_ack_packet:

                # if the last ACK packet has not been processed on the chip,
                # process it now
                last_sent_ack_sdp_packet = \
                    self._received_data.last_sent_packet_to_core(
                        placement.x, placement.y, placement.p)
                last_sent_ack_packet = \
                    create_eieio_command.read_eieio_command_message(
                        last_sent_ack_sdp_packet.data, 0)
                if not isinstance(last_sent_ack_packet, HostDataRead):
                    raise Exception(
                        "Something somewhere went terribly wrong - "
                        "I was looking for a HostDataRead packet, "
                        "while I got {0:s}".format(last_sent_ack_packet))
                for i in xrange(last_sent_ack_packet.n_requests):

                    last_ack_packet_is_of_this_region = \
                        recording_region_id == \
                        last_sent_ack_packet.region_id(i)

                    if (last_ack_packet_is_of_this_region and
                            not end_state.is_state_updated):
                        read_ptr += last_sent_ack_packet.space_read(i)
                        if (read_ptr == write_ptr or
                                (read_ptr == end_ptr and
                                 write_ptr == start_ptr)):
                            end_state.update_last_operation(
                                BUFFERING_OPERATIONS.BUFFER_READ.value)
                        if read_ptr == end_ptr:
                            read_ptr = start_ptr
                        elif read_ptr > end_ptr:
                            raise Exception(
                                "Something somewhere went terribly wrong - "
                                "I was reading beyond the region area some "
                                "unknown data".format(
                                    last_sent_ack_packet))
                end_state.update_read_pointer(read_ptr)
                end_state.set_update_completed()

            # now state is updated, read back values for read pointer and
            # last operation performed
            last_operation = end_state.last_buffer_operation
            read_ptr = end_state.current_read

            # now read_ptr is updated, check memory to read
            if read_ptr < write_ptr:
                length = write_ptr - read_ptr
                data = self._transceiver.read_memory(
                    placement.x, placement.y, read_ptr, length)
                self._received_data.flushing_data_from_region(
                    placement.x, placement.y, placement.p, recording_region_id,
                    data)

            elif read_ptr > write_ptr:
                length = end_ptr - read_ptr
                if length < 0:
                    raise exceptions.ConfigurationException(
                        "The amount of data to read is negative!")
                data = self._transceiver.read_memory(
                    placement.x, placement.y, read_ptr, length)
                self._received_data.store_data_in_region_buffer(
                    placement.x, placement.y, placement.p, recording_region_id,
                    data)
                read_ptr = start_ptr
                length = write_ptr - read_ptr
                data = self._transceiver.read_memory(
                    placement.x, placement.y, read_ptr, length)
                self._received_data.flushing_data_from_region(
                    placement.x, placement.y, placement.p, recording_region_id,
                    data)

            elif (read_ptr == write_ptr and
                    last_operation == BUFFERING_OPERATIONS.BUFFER_WRITE.value):
                length = end_ptr - read_ptr
                data = self._transceiver.read_memory(
                    placement.x, placement.y, read_ptr, length)
                self._received_data.store_data_in_region_buffer(
                    placement.x, placement.y, placement.p, recording_region_id,
                    data)
                read_ptr = start_ptr
                length = write_ptr - read_ptr
                data = self._transceiver.read_memory(
                    placement.x, placement.y, read_ptr, length)
                self._received_data.flushing_data_from_region(
                    placement.x, placement.y, placement.p, recording_region_id,
                    data)

            elif (read_ptr == write_ptr and
                    last_operation == BUFFERING_OPERATIONS.BUFFER_READ.value):
                data = bytearray()
                self._received_data.flushing_data_from_region(
                    placement.x, placement.y, placement.p, recording_region_id,
                    data)

        # data flush has been completed - return appropriate data
        # the two returns can be exchanged - one returns data and the other
        # returns a pointer to the structure holding the data
        return self._received_data.get_region_data_pointer(
            placement.x, placement.y, placement.p, recording_region_id)

    def _retrieve_and_store_data(self, packet):
        """ Following a SpinnakerRequestReadData packet, the data stored\
           during the simulation needs to be read by the host and stored in a\
           data structure, following the specifications of buffering out\
           technique

        :param packet: SpinnakerRequestReadData packet received from the\
                SpiNNaker system
        :type packet:\
                :py:class:`spinnman.messages.eieio.command_messages.spinnaker_request_read_data.SpinnakerRequestReadData`
        :rtype: None
        """
        x = packet.x
        y = packet.y
        p = packet.p

        # check packet sequence number
        pkt_seq = packet.sequence_no
        last_pkt_seq = self._received_data.last_sequence_no_for_core(x, y, p)
        next_pkt_seq = (last_pkt_seq + 1) % 256
        if pkt_seq != next_pkt_seq:

            # this sequence number is incorrect
            # re-sent last HostDataRead packet sent
            last_packet_sent = self._received_data.last_sent_packet_to_core(
                x, y, p)
            if last_packet_sent is not None:
                self._transceiver.send_sdp_message(last_packet_sent)
            else:
                raise Exception(
                    "{}, {}, {}: Something somewhere went terribly wrong - "
                    "The packet sequence numbers have gone wrong "
                    "somewhere: the packet sent from the board "
                    "has incorrect sequence number, but the host "
                    "never sent one acknowledge".format(x, y, p))
            return

        # read data from memory, store it and create data for return ACK packet
        n_requests = packet.n_requests
        new_channel = list()
        new_region_id = list()
        new_space_read = list()
        new_n_requests = 0
        for i in xrange(n_requests):
            length = packet.space_to_be_read(i)
            if length > 0:
                new_n_requests += 1
                start_address = packet.start_address(i)
                region_id = packet.region_id(i)
                channel = packet.channel(i)
                data = self._transceiver.read_memory(
                    x, y, start_address, length)
                self._received_data.store_data_in_region_buffer(
                    x, y, p, region_id, data)
                new_channel.append(channel)
                new_region_id.append(region_id)
                new_space_read.append(length)

        # create return acknowledge packet with data stored
        ack_packet = HostDataRead(
            new_n_requests, pkt_seq, new_channel, new_region_id,
            new_space_read)
        ack_packet_data = ack_packet.bytestring

        # create SDP header and message
        return_message_header = SDPHeader(
            destination_port=SDP_PORTS.OUTPUT_BUFFERING_SDP_PORT.value,
            destination_cpu=p, destination_chip_x=x, destination_chip_y=y,
            flags=SDPFlag.REPLY_NOT_EXPECTED)
        return_message = SDPMessage(return_message_header, ack_packet_data)

        # storage of last packet received
        self._received_data.store_last_received_packet_from_core(
            x, y, p, packet)
        self._received_data.update_sequence_no_for_core(x, y, p, pkt_seq)

        # store last sent message and send to the appropriate core
        self._received_data.store_last_sent_packet_to_core(
            x, y, p, return_message)
        self._transceiver.send_sdp_message(return_message)

    @property
    def sender_vertices(self):
        """ The vertices which are buffered
        """
        return self._sender_vertices

    @property
    def reload_buffer_files(self):
        """ The file paths for each buffered region for each sender vertex
        """
        return self._reload_buffer_file_paths



from enum import Enum


class ExecutableStartType(Enum):
    """ supports starting of different types of executables
    """

    RUNNING = (
        0, "Runs immediately without waiting for barrier and then exits")
    SYNC = (
        1, "Calls spin1_start(SYNC_WAIT) and then eventually spin1_exit()")
    USES_SIMULATION_INTERFACE = (
        2,
        "Calls simulation_run() and simulation_exit() / "
        "simulation_handle_pause_resume()"),
    NO_APPLICATION = (3, "Situation where there user has supplied no "
                         "application but for some reason still wants to run")

    def __new__(cls, value, doc=""):
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, value, doc=""):
        self._value_ = value
        self.__doc__ = doc



# spinn front end common imports
from spinn_front_end_common.utility_models.live_packet_gather \
    import LivePacketGather
from spinn_front_end_common.utility_models.live_packet_gather_machine_vertex \
    import LivePacketGatherMachineVertex

# pacman imports
from pacman.model.graphs.common import Slice
from pacman.model.constraints.placer_constraints\
    import ChipAndCoreConstraint

from spinn_utilities.progress_bar import ProgressBar

from collections import defaultdict


class InsertLivePacketGatherersToGraphs(object):
    """ function to add LPG's as required into a given graph
    """

    def __call__(
            self, live_packet_gatherer_parameters, machine, machine_graph,
            application_graph=None, graph_mapper=None):
        """ call that adds LPG vertices on Ethernet connected chips as\
            required.

        :param live_packet_gatherer_parameters:\
            the Live Packet Gatherer parameters requested by the script
        :param machine: the spinnaker machine as discovered
        :param application_graph: the application graph
        :param machine_graph: the machine graph
        :return: mapping between LPG params and LPG vertex
        """

        # create progress bar
        progress = ProgressBar(
            machine.ethernet_connected_chips,
            string_describing_what_being_progressed=(
                "Adding Live Packet Gatherers to Graph"))

        # Keep track of the vertices added by parameters and board address
        lpg_params_to_vertices = defaultdict(dict)

        # for every Ethernet connected chip, add the gatherers required
        for chip in progress.over(machine.ethernet_connected_chips):
            for lpg_params in live_packet_gatherer_parameters:
                if (lpg_params.board_address is None or
                        lpg_params.board_address == chip.ip_address):
                    lpg_params_to_vertices[lpg_params][chip.x, chip.y] = \
                        self._add_lpg_vertex(application_graph, graph_mapper,
                                             machine_graph, chip, lpg_params)

        return lpg_params_to_vertices

    def _add_lpg_vertex(self, app_graph, mapper, m_graph, chip, lpg_params):
        if app_graph is not None:
            vtx_slice = Slice(0, 0)
            app_vtx = self._create_vertex(LivePacketGather, lpg_params)
            app_graph.add_vertex(app_vtx)
            resources_required = app_vtx.get_resources_used_by_atoms(
                vtx_slice)
            m_vtx = app_vtx.create_machine_vertex(
                vtx_slice, resources_required)
            mapper.add_vertex_mapping(m_vtx, vtx_slice, app_vtx)
        else:
            m_vtx = self._create_vertex(
                LivePacketGatherMachineVertex, lpg_params)

        m_vtx.add_constraint(ChipAndCoreConstraint(x=chip.x, y=chip.y))
        m_graph.add_vertex(m_vtx)
        return m_vtx

    @staticmethod
    def _create_vertex(lpg_vertex_class, params):
        """ Creates a Live Packet Gather Vertex

        :param lpg_vertex_class: the type to create for the vertex
        :param params: the params of the vertex
        :return the vertex built
        """
        return lpg_vertex_class(
            hostname=params.hostname,
            port=params.port,
            tag=params.tag,
            board_address=params.board_address,
            strip_sdp=params.strip_sdp,
            use_prefix=params.use_prefix,
            key_prefix=params.key_prefix,
            prefix_type=params.prefix_type,
            message_type=params.message_type,
            right_shift=params.right_shift,
            payload_as_time_stamps=params.payload_as_time_stamps,
            use_payload_prefix=params.use_payload_prefix,
            payload_prefix=params.payload_prefix,
            payload_right_shift=params.payload_right_shift,
            number_of_packets_sent_per_time_step=(
                params.number_of_packets_sent_per_time_step),
            label="LiveSpikeReceiver")



from spinn_front_end_common.utilities.utility_objs import ProvenanceDataItem


class PacmanProvenanceExtractor(object):
    """ Extracts Provenance data from a PACMANAlgorithmExecutor
    """

    def __init__(self):
        self._data_items = list()

    def extract_provenance(self, executor):
        """ acquires the timings from pacman algorithms (provenance data)

        :param executor: the pacman workflow executor
        :rtype: None
        """
        for (algorithm, run_time, exec_names) in executor.algorithm_timings:
            names = ["pacman"]
            names.append(exec_names)
            names.extend(["run_time_of_{}".format(algorithm)])
            self._data_items.append(ProvenanceDataItem(names, run_time))

    @property
    def data_items(self):
        """ returns the provenance data items

        :return: list of provenance data items.
        :rtype: iterable of ProvenanceDataItem
        """
        return self._data_items

    def clear(self):
        """ clears the provenance data store

        :rtype: None
        """
        self._data_items = list()



from pacman.executor.injection_decorator import inject_items
from pacman.model.graphs.application import ApplicationVertex

from spinn_front_end_common.abstract_models.\
    abstract_generates_data_specification import \
    AbstractGeneratesDataSpecification
from spinn_front_end_common.abstract_models.\
    abstract_has_associated_binary import \
    AbstractHasAssociatedBinary
from spinn_front_end_common.utility_models.\
    chip_power_monitor_machine_vertex import ChipPowerMonitorMachineVertex

from spinn_utilities.overrides import overrides


class ChipPowerMonitorApplicationVertex(
        ApplicationVertex, AbstractHasAssociatedBinary,
        AbstractGeneratesDataSpecification):
    """ class for representing idle time recording code in a application graph.
    """

    def __init__(
            self, label, constraints, n_samples_per_recording,
            sampling_frequency):
        """ chip power monitor application vertex constructor

        :param label: vertex label
        :param constraints: constraints for the vertex
        :param n_samples_per_recording: how many samples to take before /
            recording to sdram the total
        :param sampling_frequency: how many microseconds between sampling
        """
        ApplicationVertex.__init__(self, label, constraints, 1)
        self._n_samples_per_recording = n_samples_per_recording
        self._sampling_frequency = sampling_frequency

    @property
    @overrides(ApplicationVertex.n_atoms)
    def n_atoms(self):
        return 1

    @overrides(ApplicationVertex.create_machine_vertex)
    def create_machine_vertex(self, vertex_slice, resources_required,
                              label=None, constraints=None):
        return ChipPowerMonitorMachineVertex(
            constraints=constraints, label=label,
            n_samples_per_recording=self._n_samples_per_recording,
            sampling_frequency=self._sampling_frequency)

    @overrides(AbstractHasAssociatedBinary.get_binary_file_name)
    def get_binary_file_name(self):
        return ChipPowerMonitorMachineVertex.binary_file_name()

    @inject_items({"time_scale_factor": "TimeScaleFactor",
                   "machine_time_step": "MachineTimeStep",
                   "n_machine_time_steps": "TotalMachineTimeSteps",
                   "ip_tags": "MemoryIpTags"})
    @overrides(
        AbstractGeneratesDataSpecification.generate_data_specification,
        additional_arguments={
            "machine_time_step", "time_scale_factor", "n_machine_time_steps",
            "ip_tags"})
    def generate_data_specification(
            self, spec, placement, machine_time_step, time_scale_factor,
            n_machine_time_steps, ip_tags):
        # generate spec for the machine vertex
        placement.vertex._generate_data_specification(
            spec, machine_time_step, time_scale_factor, n_machine_time_steps,
            ip_tags)

    @overrides(AbstractHasAssociatedBinary.get_binary_start_type)
    def get_binary_start_type(self):
        return ChipPowerMonitorMachineVertex.binary_start_type()

    @overrides(ApplicationVertex.get_resources_used_by_atoms)
    def get_resources_used_by_atoms(self, vertex_slice):
        return self._get_resources_used_by_atoms(vertex_slice)

    @inject_items({
        "n_machine_time_steps": "TotalMachineTimeSteps",
        "machine_time_step": "MachineTimeStep",
        "time_scale_factor": "TimeScaleFactor"})
    def _get_resources_used_by_atoms(
            self, vertex_slice, n_machine_time_steps, time_scale_factor,
            machine_time_step):
        return ChipPowerMonitorMachineVertex.get_resources(
            n_machine_time_steps, machine_time_step, time_scale_factor,
            self._n_samples_per_recording, self._sampling_frequency)



from .memory_map_on_chip_report import MemoryMapOnChipReport
from .memory_map_on_host_chip_report import MemoryMapOnHostChipReport
from .memory_map_on_host_report import MemoryMapOnHostReport
from .routing_table_from_machine_report import RoutingTableFromMachineReport

__all__ = [
    "MemoryMapOnChipReport",
    "MemoryMapOnHostChipReport",
    "MemoryMapOnHostReport",
    "RoutingTableFromMachineReport"
]



from .command_sender import CommandSender
from .command_sender_machine_vertex import CommandSenderMachineVertex
from .live_packet_gather import LivePacketGather
from .live_packet_gather_machine_vertex import LivePacketGatherMachineVertex
from .multi_cast_command import MultiCastCommand
from .reverse_ip_tag_multi_cast_source import ReverseIpTagMultiCastSource
from .reverse_ip_tag_multicast_source_machine_vertex \
    import ReverseIPTagMulticastSourceMachineVertex

__all__ = ["CommandSender", "CommandSenderMachineVertex",
           "LivePacketGather", "LivePacketGatherMachineVertex",
           "MultiCastCommand", "ReverseIpTagMultiCastSource",
           "ReverseIPTagMulticastSourceMachineVertex"]



from collections import defaultdict
from spinn_storage_handlers \
    import BufferedBytearrayDataStorage, BufferedTempfileDataStorage


class BufferedReceivingData(object):
    """ Stores the information received through the buffering output\
        technique from the SpiNNaker system. The data kept includes the last\
        sent packet and last received packet, their correspondent sequence\
        numbers, the data retrieved, a flag to identify if the data from a\
        core has been flushed and the final state of the buffering output\
        state machine
    """

    __slots__ = [

        # the data to store
        "_data",

        # dict of booleans indicating if a region on a core has been flushed
        "_is_flushed",

        # dict of last sequence number received by core
        "_sequence_no",

        # dict of last packet received by core
        "_last_packet_received",

        # dict of last packet sent by core
        "_last_packet_sent",

        # dict of end buffer sequence number
        "_end_buffering_sequence_no",

        # dict of end state by core
        "_end_buffering_state"
    ]

    def __init__(self, store_to_file=False):
        """

        :param store_to_file: A boolean to identify if the data will be stored\
                in memory using a byte array or in a temporary file on the disk
        :type store_to_file: bool
        """

        self._data = None
        if store_to_file:
            self._data = defaultdict(BufferedTempfileDataStorage)
        else:
            self._data = defaultdict(BufferedBytearrayDataStorage)
        self._is_flushed = defaultdict(lambda: False)
        self._sequence_no = defaultdict(lambda: 0xFF)
        self._last_packet_received = defaultdict(lambda: None)
        self._last_packet_sent = defaultdict(lambda: None)
        self._end_buffering_sequence_no = dict()
        self._end_buffering_state = dict()

    def store_data_in_region_buffer(self, x, y, p, region, data):
        """ Store some information in the correspondent buffer class for a\
            specific chip, core and region

        :param x: x coordinate of the chip
        :type x: int
        :param y: y coordinate of the chip
        :type y: int
        :param p: Core within the specified chip
        :type p: int
        :param region: Region containing the data to be stored
        :type region: int
        :param data: data to be stored
        :type data: bytearray
        """
        self._data[x, y, p, region].write(data)

    def is_data_from_region_flushed(self, x, y, p, region):
        """ Check if the data region has been flushed

        :param x: x coordinate of the chip
        :type x: int
        :param y: y coordinate of the chip
        :type y: int
        :param p: Core within the specified chip
        :type p: int
        :param region: Region containing the data
        :type region: int
        :return: True if the region has been flushed. False otherwise
        :rtype: bool
        """
        return self._is_flushed[x, y, p, region]

    def flushing_data_from_region(self, x, y, p, region, data):
        """ Store flushed data from a region of a core on a chip, and mark it\
            as being flushed

        :param x: x coordinate of the chip
        :type x: int
        :param y: y coordinate of the chip
        :type y: int
        :param p: Core within the specified chip
        :type p: int
        :param region: Region containing the data to be stored
        :type region: int
        :param data: data to be stored
        :type data: bytearray
        """
        self.store_data_in_region_buffer(x, y, p, region, data)
        self._is_flushed[x, y, p, region] = True

    def store_last_received_packet_from_core(self, x, y, p, packet):
        """ Store the most recent packet received from SpiNNaker for a given\
            core

        :param x: x coordinate of the chip
        :type x: int
        :param y: y coordinate of the chip
        :type y: int
        :param p: Core within the specified chip
        :type p: int
        :param packet: SpinnakerRequestReadData packet received
        :type packet:\
                :py:class:`spinnman.messages.eieio.command_messages.spinnaker_request_read_data.SpinnakerRequestReadData`
        """
        self._last_packet_received[x, y, p] = packet

    def last_received_packet_from_core(self, x, y, p):
        """ Get the last packet received for a given core

        :param x: x coordinate of the chip
        :type x: int
        :param y: y coordinate of the chip
        :type y: int
        :param p: Core within the specified chip
        :type p: int
        :return: SpinnakerRequestReadData packet received
        :rtype:\
                :py:class:`spinnman.messages.eieio.command_messages.spinnaker_request_read_data.SpinnakerRequestReadData`
        """
        return self._last_packet_received[x, y, p]

    def store_last_sent_packet_to_core(self, x, y, p, packet):
        """ Store the last packet sent to the given core

        :param x: x coordinate of the chip
        :type x: int
        :param y: y coordinate of the chip
        :type y: int
        :param p: Core within the specified chip
        :type p: int
        :param packet: last HostDataRead packet sent
        :type packet:\
                :py:class:`spinnman.messages.eieio.command_messages.host_data_read.HostDataRead`
        """
        self._last_packet_sent[x, y, p] = packet

    def last_sent_packet_to_core(self, x, y, p):
        """ Retrieve the last packet sent to a core

        :param x: x coordinate of the chip
        :type x: int
        :param y: y coordinate of the chip
        :type y: int
        :param p: Core within the specified chip
        :type p: int
        :return: last HostDataRead packet sent
        :rtype:\
                :py:class:`spinnman.messages.eieio.command_messages.host_data_read.HostDataRead`
        """
        return self._last_packet_sent[x, y, p]

    def last_sequence_no_for_core(self, x, y, p):
        """ Get the last sequence number for a core

        :param x: x coordinate of the chip
        :type x: int
        :param y: y coordinate of the chip
        :type y: int
        :param p: Core within the specified chip
        :type p: int
        :return: last sequence number used
        :rtype: int
        """
        return self._sequence_no[x, y, p]

    def update_sequence_no_for_core(self, x, y, p, sequence_no):
        """ Set the last sequence number used

        :param x: x coordinate of the chip
        :type x: int
        :param y: y coordinate of the chip
        :type y: int
        :param p: Core within the specified chip
        :type p: int
        :param sequence_no: last sequence number used
        :type sequence_no: int
        :rtype: None
        """
        self._sequence_no[x, y, p] = sequence_no

    def get_region_data(self, x, y, p, region):
        """ Get the data stored for a given region of a given core

        :param x: x coordinate of the chip
        :type x: int
        :param y: y coordinate of the chip
        :type y: int
        :param p: Core within the specified chip
        :type p: int
        :param region: Region containing the data
        :type region: int
        :return: an array contained all the data received during the\,
                simulation, and a flag indicating if any data was missing
        :rtype: (bytearray, bool)
        """
        missing = None
        if (x, y, p, region) not in self._end_buffering_state:
            missing = (x, y, p, region)
        data = self._data[x, y, p, region].read_all()
        return data, missing

    def get_region_data_pointer(self, x, y, p, region):
        """ Get the data received during the simulation for a region of a core

        :param x: x coordinate of the chip
        :type x: int
        :param y: y coordinate of the chip
        :type y: int
        :param p: Core within the specified chip
        :type p: int
        :param region: Region containing the data
        :type region: int
        :return: all the data received during the simulation,\
                and a flag indicating if any data was lost
        :rtype:\
            (:py:class:`spinn_front_end_common.interface.buffer_management.buffer_models.AbstractBufferedDataStorage`,
             bool)
        """
        missing = False
        if (x, y, p, region) not in self._end_buffering_state:
            missing = True
        data_pointer = self._data[x, y, p, region]
        return data_pointer, missing

    def store_end_buffering_state(self, x, y, p, region, state):
        """ Store the end state of buffering

        :param x: x coordinate of the chip
        :type x: int
        :param y: y coordinate of the chip
        :type y: int
        :param p: Core within the specified chip
        :type p: int
        :param state: The end state
        """
        self._end_buffering_state[x, y, p, region] = state

    def is_end_buffering_state_recovered(self, x, y, p, region):
        """ Determine if the end state has been stored

        :param x: x coordinate of the chip
        :type x: int
        :param y: y coordinate of the chip
        :type y: int
        :param p: Core within the specified chip
        :type p: int
        :return: True if the state has been stored
        """
        return (x, y, p, region) in self._end_buffering_state

    def get_end_buffering_state(self, x, y, p, region):
        """ Get the end state of the buffering

        :param x: x coordinate of the chip
        :type x: int
        :param y: y coordinate of the chip
        :type y: int
        :param p: Core within the specified chip
        :type p: int
        :return: The end state
        """
        return self._end_buffering_state[x, y, p, region]

    def store_end_buffering_sequence_number(self, x, y, p, sequence):
        """ Store the last sequence number sent by the core

        :param x: x coordinate of the chip
        :type x: int
        :param y: y coordinate of the chip
        :type y: int
        :param p: Core within the specified chip
        :type p: int
        :param sequence: The last sequence number
        :type sequence: int
        """
        self._end_buffering_sequence_no[x, y, p] = sequence

    def is_end_buffering_sequence_number_stored(self, x, y, p):
        """ Determine if the last sequence number has been retrieved

        :param x: x coordinate of the chip
        :type x: int
        :param y: y coordinate of the chip
        :type y: int
        :param p: Core within the specified chip
        :type p: int
        :return: True if the number has been retrieved
        :rtype: bool
        """
        return (x, y, p) in self._end_buffering_sequence_no

    def get_end_buffering_sequence_number(self, x, y, p):
        """ Get the last sequence number sent by the core

        :param x: x coordinate of the chip
        :type x: int
        :param y: y coordinate of the chip
        :type y: int
        :param p: Core within the specified chip
        :type p: int
        :return: The last sequence number
        :rtype: int
        """
        return self._end_buffering_sequence_no[x, y, p]

    def resume(self):
        """ Resets states so that it can behave in a resumed mode
        """
        self._end_buffering_state = dict()
        self._is_flushed = defaultdict(lambda: False)
        self._sequence_no = defaultdict(lambda: 0xFF)
        self._last_packet_received = defaultdict(lambda: None)
        self._last_packet_sent = defaultdict(lambda: None)

    def clear(self, x, y, p, region_id):
        """ clears the data from a given data region (only clears things\
            associated with a given data recording region).

        :param x: placement x coord
        :param y: placement y coord
        :param p: placement p coord
        :param region_id: the recording region id to clear data from
        :rtype: None
        """
        del self._end_buffering_state[x, y, p, region_id]
        del self._data[x, y, p, region_id]
        del self._is_flushed[x, y, p, region_id]



import os
from spinn_utilities.executable_finder import ExecutableFinder as BaseEF
from spinn_front_end_common import common_model_binaries


class ExecutableFinder(BaseEF):
    """ Manages a set of folders in which to search for binaries,
        and allows for binaries to be discovered within this path.
        This adds a default location to look to the base class.
    """

    def __init__(self, binary_search_paths=None,
                 include_common_binaries_folder=True):
        """

        :param binary_search_paths: The initial set of folders to search for\
                    binaries.
        :type binary_search_paths: iterable of str
        :param include_common_binaries_folder: If True (i.e. the default), \
                    the spinn_front_end_common.common_model_binaries folder \
                    is searched for binaries.  If you are not using the common\
                    models, or the model binary names conflict with your own,\
                    this parameter can be used to avoid this issue.  Note that\
                    the folder will be appended to the value of\
                    binary_search_paths, so the common binary search path will\
                    be looked in last.
        :type include_common_binaries_folder: bool
        """
        if binary_search_paths is None:
            binary_search_paths = list()
        BaseEF.__init__(self, binary_search_paths)
        if include_common_binaries_folder:
            self.add_path(os.path.dirname(
                common_model_binaries.__file__))



import struct
import logging

from spinn_utilities.progress_bar import ProgressBar
from spinnman.messages.sdp import SDPFlag, SDPHeader, SDPMessage
from spinnman.model.enums import CPUState
from spinn_front_end_common.utilities.constants \
    import SDP_PORTS, SDP_RUNNING_MESSAGE_CODES
from spinn_front_end_common.utilities.exceptions import ConfigurationException

logger = logging.getLogger(__name__)


class ChipProvenanceUpdater(object):
    """ Forces all cores to generate provenance data, and then exit
    """

    __slots__ = []

    def __call__(self, txrx, app_id, all_core_subsets):
        # check that the right number of processors are in sync
        processors_completed = txrx.get_core_state_count(
            app_id, CPUState.FINISHED)
        total_processors = len(all_core_subsets)
        left_to_do_cores = total_processors - processors_completed

        progress = ProgressBar(
            left_to_do_cores,
            "Forcing error cores to generate provenance data")

        error_cores = txrx.get_cores_in_state(
            all_core_subsets, CPUState.RUN_TIME_EXCEPTION)
        watchdog_cores = txrx.get_cores_in_state(
            all_core_subsets, CPUState.WATCHDOG)
        idle_cores = txrx.get_cores_in_state(
            all_core_subsets, CPUState.IDLE)

        if (len(error_cores) != 0 or len(watchdog_cores) != 0 or
                len(idle_cores) != 0):
            raise ConfigurationException(
                "Some cores have crashed. RTE cores {}, watch-dogged cores {},"
                " idle cores {}".format(
                    error_cores.values(), watchdog_cores.values(),
                    idle_cores.values()))

        # check that all cores are in the state FINISHED which shows that
        # the core has received the message and done provenance updating
        attempts = 0
        while processors_completed != total_processors and attempts < 10:
            attempts += 1
            unsuccessful_cores = txrx.get_cores_not_in_state(
                all_core_subsets, CPUState.FINISHED)

            for (x, y, p) in unsuccessful_cores.iterkeys():
                data = struct.pack(
                    "<I", SDP_RUNNING_MESSAGE_CODES.
                    SDP_UPDATE_PROVENCE_REGION_AND_EXIT.value)
                txrx.send_sdp_message(SDPMessage(SDPHeader(
                    flags=SDPFlag.REPLY_NOT_EXPECTED,
                    destination_cpu=p,
                    destination_chip_x=x,
                    destination_port=SDP_PORTS.RUNNING_COMMAND_SDP_PORT.value,
                    destination_chip_y=y), data=data))

            processors_completed = txrx.get_core_state_count(
                app_id, CPUState.FINISHED)

            left_over_now = total_processors - processors_completed
            to_update = left_to_do_cores - left_over_now
            left_to_do_cores = left_over_now
            if to_update != 0:
                progress.update(to_update)
        if attempts >= 10:
            logger.error("Unable to Finish getting provenance data. "
                         "Abandoned after too many retries. "
                         "Board may be left in an unstable state!")

        progress.end()



from spinn_utilities.progress_bar import ProgressBar

from .scp_clear_iobuf_request import SCPClearIOBUFRequest
from spinn_front_end_common.utilities.constants import SDP_PORTS
from spinnman.processes import AbstractMultiConnectionProcess


class ClearIOBUFProcess(AbstractMultiConnectionProcess):

    def __init__(self, connection_selector):
        AbstractMultiConnectionProcess.__init__(self, connection_selector)
        self._progress = None

    def receive_response(self, response):  # @UnusedVariable
        if self._progress is not None:
            self._progress.update()

    def clear_iobuf(self, core_subsets, n_cores):
        self._progress = ProgressBar(
            n_cores, "clearing IOBUF from the machine")
        for core_subset in core_subsets:
            for processor_id in core_subset.processor_ids:
                self._send_request(
                    SCPClearIOBUFRequest(
                        core_subset.x, core_subset.y, processor_id,
                        SDP_PORTS.RUNNING_COMMAND_SDP_PORT.value),
                    callback=self.receive_response)
        self._finish()
        self._progress.end()
        self.check_for_error()



from spalloc import ProtocolClient


class SpallocMaxMachineGenerator(object):
    """ Generates the width and height of the maximum machine a given\
        allocation server can generate
    """

    __slots__ = []

    def __call__(
            self, spalloc_server, spalloc_port=22244, spalloc_machine=None):
        with ProtocolClient(spalloc_server, spalloc_port) as client:
            machines = client.list_machines()
            # Close the context immediately; don't want to keep this particular
            # connection around as there's not a great chance of this code
            # being rerun in this process any time soon.
        max_width = None
        max_height = None
        max_area = -1

        for machine in self._filter(machines, spalloc_machine):
            # Get the width and height in chips, and logical area in chips**2
            width, height, area = self._get_size(machine)

            # The "biggest" board is the one with the most chips
            if area > max_area:
                max_area = area
                max_width = width
                max_height = height

        # Return the width and height, and make no assumption about wrap-
        # arounds or version.
        return max_width, max_height, None, None

    @staticmethod
    def _filter(machines, filter):  # @ReservedAssignment
        if filter is None:
            return (m for m in machines if "default" in m["tags"])
        else:
            return (m for m in machines if m["name"] == filter)

    @staticmethod
    def _get_size(machine):
        # Get the width and height in chips
        width = machine["width"] * 12
        height = machine["height"] * 12

        # A specific exception is the 1-board machine, which is represented as
        # a 3 board machine with 2 dead boards. In this case the width and
        # height are 8.
        if (machine["width"] == 1 and
                machine["height"] == 1 and
                len(machine["dead_boards"]) == 2):
            width = 8
            height = 8

        return width, height, width * height



from spinn_utilities.progress_bar import ProgressBar

# front end common imports
from spinn_front_end_common.interface.provenance \
    import AbstractProvidesLocalProvenanceData


class GraphProvenanceGatherer(object):
    __slots__ = []

    def __call__(
            self, machine_graph, application_graph=None,
            provenance_data_objects=None):
        """

        :param machine_graph: The machine graph to inspect
        :param application_graph: The optional application graph
        :param provenance_data_objects: Any existing objects to append to
        """

        if provenance_data_objects is not None:
            prov_items = provenance_data_objects
        else:
            prov_items = list()

        progress = ProgressBar(
            machine_graph.n_vertices +
            machine_graph.n_outgoing_edge_partitions,
            "Getting provenance data from machine graph")
        for vertex in machine_graph.vertices:
            if isinstance(vertex, AbstractProvidesLocalProvenanceData):
                prov_items.extend(vertex.get_local_provenance_data())
            progress.update()
        for partition in machine_graph.outgoing_edge_partitions:
            for edge in partition.edges:
                if isinstance(edge, AbstractProvidesLocalProvenanceData):
                    prov_items.extend(edge.get_local_provenance_data())
            progress.update()
        progress.end()

        if application_graph is not None:
            progress = ProgressBar(
                application_graph.n_vertices +
                application_graph.n_outgoing_edge_partitions,
                "Getting provenance data from application graph")
            for vertex in application_graph.vertices:
                if isinstance(vertex, AbstractProvidesLocalProvenanceData):
                    prov_items.extend(vertex.get_local_provenance_data())
                progress.update()
            for partition in application_graph.outgoing_edge_partitions:
                for edge in partition.edges:
                    if isinstance(edge, AbstractProvidesLocalProvenanceData):
                        prov_items.extend(edge.get_local_provenance_data())
                progress.update()
            progress.end()

        return prov_items



from six import add_metaclass
from spinn_utilities.abstract_base import AbstractBase
from spinn_utilities.abstract_base import abstractmethod


@add_metaclass(AbstractBase)
class AbstractHasProfileData(object):
    """ Indicates an object that can record a profile
    """

    @abstractmethod
    def get_profile_data(self, transceiver, placement):
        """ Get the profile data recorded during simulation

        :rtype:\
            :py:class:`spinn_front_end_common.interface.profiling.profile_data.ProfileData`
        """



from spinn_utilities.progress_bar import ProgressBar
from spinn_machine import SDRAM

from data_specification import DataSpecificationExecutor
from data_specification import utility_calls
from data_specification.constants import MAX_MEM_REGIONS

from spinn_front_end_common.abstract_models \
    import AbstractRewritesDataSpecification
from spinn_front_end_common.utilities import helpful_functions

from spinn_storage_handlers import FileDataReader

import os
import struct


class DSGRegionReloader(object):
    """ Regenerates Data Specifications
    """

    def __call__(
            self, transceiver, placements, hostname, report_directory,
            write_text_specs, application_data_file_path, graph_mapper=None):
        """

        :param transceiver: SpiNNMan transceiver for communication
        :param placements: the list of placements of the machine graph to cores
        :param hostname: the machine name
        :param report_directory: the location where reports are stored
        :param write_text_specs:\
            True if the textual version of the specification is to be written
        :param application_data_file_path:\
            Folder where data specifications should be written to
        :param graph_mapper:\
            the mapping between application and machine graph
        """

        # build file paths for reloaded stuff
        reloaded_dsg_data_files_file_path = \
            helpful_functions.generate_unique_folder_name(
                application_data_file_path, "reloaded_data_regions", "")
        reloaded_dsg_report_files_file_path = \
            helpful_functions.generate_unique_folder_name(
                report_directory, "reloaded_data_regions", "")

        # build new folders
        if not os.path.exists(reloaded_dsg_data_files_file_path):
            os.makedirs(reloaded_dsg_data_files_file_path)
        if not os.path.exists(reloaded_dsg_report_files_file_path):
            os.makedirs(reloaded_dsg_report_files_file_path)

        application_vertices_to_reset = set()

        progress = ProgressBar(placements.n_placements, "Reloading data")
        for placement in progress.over(placements.placements):

            # Try to generate the data spec for the placement
            generated = self._regenerate_data_spec_for_vertices(
                transceiver, placement, placement.vertex, hostname,
                reloaded_dsg_report_files_file_path, write_text_specs,
                reloaded_dsg_data_files_file_path)

            # If the region was regenerated, mark it reloaded
            if generated:
                placement.vertex.mark_regions_reloaded()

            # If the spec wasn't generated directly, and there is an
            # application vertex, try with that
            if not generated and graph_mapper is not None:
                associated_vertex = graph_mapper.get_application_vertex(
                    placement.vertex)
                generated = self._regenerate_data_spec_for_vertices(
                    transceiver, placement, associated_vertex, hostname,
                    reloaded_dsg_report_files_file_path, write_text_specs,
                    reloaded_dsg_data_files_file_path)

                # If the region was regenerated, remember the application
                # vertex for resetting later
                if generated:
                    application_vertices_to_reset.add(associated_vertex)

        # Only reset the application vertices here, otherwise only one
        # machine vertices data will be updated
        for vertex in application_vertices_to_reset:
            vertex.mark_regions_reloaded()

    @staticmethod
    def _regenerate_data_spec_for_vertices(
            transceiver, placement, vertex, hostname,
            reloaded_dsg_report_files_file_path, write_text_specs,
            reloaded_dsg_data_files_file_path):

        # If the vertex doesn't regenerate, skip
        if not isinstance(vertex, AbstractRewritesDataSpecification):
            return False

        # If the vertex doesn't require regeneration, skip
        if not vertex.requires_memory_regions_to_be_reloaded():
            return True

        # build the writers for the reports and data
        spec_file, spec = utility_calls.get_data_spec_and_file_writer_filename(
            placement.x, placement.y, placement.p, hostname,
            reloaded_dsg_report_files_file_path,
            write_text_specs, reloaded_dsg_data_files_file_path)

        # Execute the regeneration
        vertex.regenerate_data_specification(spec, placement)

        # execute the spec
        spec_reader = FileDataReader(spec_file)
        data_spec_executor = DataSpecificationExecutor(
            spec_reader, SDRAM.DEFAULT_SDRAM_BYTES)
        data_spec_executor.execute()

        # Read the region table for the placement
        regions_base_address = transceiver.get_cpu_information_from_core(
            placement.x, placement.y, placement.p).user[0]
        start_region = utility_calls.get_region_base_address_offset(
            regions_base_address, 0)
        table_size = utility_calls.get_region_base_address_offset(
            regions_base_address, MAX_MEM_REGIONS) - start_region
        offsets = struct.unpack_from(
            "<{}I".format(MAX_MEM_REGIONS),
            transceiver.read_memory(
                placement.x, placement.y, start_region, table_size))

        # Write the regions to the machine
        for i, region in enumerate(data_spec_executor.dsef.mem_regions):
            if region is not None and not region.unfilled:
                transceiver.write_memory(
                    placement.x, placement.y, offsets[i],
                    region.region_data[:region.max_write_pointer])

        return True



from .abstract_receive_buffers_to_host import AbstractReceiveBuffersToHost
from .abstract_sends_buffers_from_host import AbstractSendsBuffersFromHost
from .sends_buffers_from_host_pre_buffered_impl \
    import SendsBuffersFromHostPreBufferedImpl

__all__ = ["AbstractReceiveBuffersToHost", "AbstractSendsBuffersFromHost",
           "SendsBuffersFromHostPreBufferedImpl"]



from six import add_metaclass

from spinn_utilities.abstract_base import AbstractBase, abstractproperty

from .abstract_provides_provenance_data_from_machine \
    import AbstractProvidesProvenanceDataFromMachine
from spinn_front_end_common.utilities.utility_objs import ProvenanceDataItem

from data_specification.utility_calls import get_region_base_address_offset

import struct
from enum import Enum


@add_metaclass(AbstractBase)
class ProvidesProvenanceDataFromMachineImpl(
        AbstractProvidesProvenanceDataFromMachine):
    """ An implementation that gets provenance data from a region of ints on\
        the machine
    """

    __slots__ = ()

    # entries for the provenance data generated by models using provides
    # provenance vertex
    PROVENANCE_DATA_ENTRIES = Enum(
        value="PROVENANCE_DATA_ENTRIES",
        names=[("TRANSMISSION_EVENT_OVERFLOW", 0),
               ("CALLBACK_QUEUE_OVERLOADED", 1),
               ("DMA_QUEUE_OVERLOADED", 2),
               ("TIMER_TIC_HAS_OVERRUN", 3),
               ("MAX_NUMBER_OF_TIMER_TIC_OVERRUN", 4)]
    )

    @abstractproperty
    def _provenance_region_id(self):
        """

        :return: provenance_region_id
        """

    @abstractproperty
    def _n_additional_data_items(self):
        """

        :return: n_additional_data_items
        """

    def reserve_provenance_data_region(self, spec):
        spec.reserve_memory_region(
            self._provenance_region_id,
            (len(self.PROVENANCE_DATA_ENTRIES) +
                self._n_additional_data_items) * 4,
            label="Provenance", empty=True)

    @staticmethod
    def get_provenance_data_size(n_additional_data_items):
        return (
            (len(ProvidesProvenanceDataFromMachineImpl
                 .PROVENANCE_DATA_ENTRIES) + n_additional_data_items) * 4)

    def _get_provenance_region_address(self, transceiver, placement):

        # Get the App Data for the core
        app_data_base_address = transceiver.get_cpu_information_from_core(
            placement.x, placement.y, placement.p).user[0]

        # Get the provenance region base address
        base_address_offset = get_region_base_address_offset(
            app_data_base_address, self._provenance_region_id)
        base_address_buffer = buffer(transceiver.read_memory(
            placement.x, placement.y, base_address_offset, 4))
        return struct.unpack("<I", base_address_buffer)[0]

    def _read_provenance_data(self, transceiver, placement):
        provenance_address = self._get_provenance_region_address(
            transceiver, placement)
        data = buffer(transceiver.read_memory(
            placement.x, placement.y, provenance_address,
            self.get_provenance_data_size(self._n_additional_data_items)))
        return struct.unpack_from("<{}I".format(
            len(self.PROVENANCE_DATA_ENTRIES) + self._n_additional_data_items),
            data)

    @staticmethod
    def _get_placement_details(placement):
        label = placement.vertex.label
        x = placement.x
        y = placement.y
        p = placement.p
        names = ["{}_{}_{}_{}".format(x, y, p, label)]
        return label, x, y, p, names

    @staticmethod
    def _add_name(names, name):
        new_names = list(names)
        new_names.append(name)
        return new_names

    @staticmethod
    def _add_names(names, extra_names):
        new_names = list(names)
        new_names.extend(extra_names)
        return new_names

    def _read_basic_provenance_items(self, provenance_data, placement):
        transmission_event_overflow = provenance_data[
            self.PROVENANCE_DATA_ENTRIES.TRANSMISSION_EVENT_OVERFLOW.value]
        callback_queue_overloaded = provenance_data[
            self.PROVENANCE_DATA_ENTRIES.CALLBACK_QUEUE_OVERLOADED.value]
        dma_queue_overloaded = provenance_data[
            self.PROVENANCE_DATA_ENTRIES.DMA_QUEUE_OVERLOADED.value]
        number_of_times_timer_tic_over_ran = provenance_data[
            self.PROVENANCE_DATA_ENTRIES.TIMER_TIC_HAS_OVERRUN.value]
        max_number_of_times_timer_tic_over_ran = provenance_data[
            self.PROVENANCE_DATA_ENTRIES.MAX_NUMBER_OF_TIMER_TIC_OVERRUN.value]

        # create provenance data items for returning
        label, x, y, p, names = self._get_placement_details(placement)
        data_items = list()
        data_items.append(ProvenanceDataItem(
            self._add_name(names, "Times_the_transmission_of_spikes_overran"),
            transmission_event_overflow,
            report=transmission_event_overflow != 0,
            message=(
                "The transmission buffer for {} on {}, {}, {} was blocked "
                "on {} occasions. This is often a sign that the system is "
                "experiencing back pressure from the communication fabric. "
                "Please either: "
                "1. spread the load over more cores, "
                "2. reduce your peak transmission load,"
                "3. adjust your mapping algorithm.".format(
                    label, x, y, p, transmission_event_overflow))))

        data_items.append(ProvenanceDataItem(
            self._add_name(names, "Times_the_callback_queue_was_overloaded"),
            callback_queue_overloaded,
            report=callback_queue_overloaded != 0,
            message=(
                "The callback queue for {} on {}, {}, {} overloaded on {} "
                "occasions. This is often a sign that the system is running "
                "too quickly for the number of neurons per core.  Please "
                "increase the machine time step or time_scale_factor or "
                "decrease the number of neurons per core.".format(
                    label, x, y, p, callback_queue_overloaded))))

        data_items.append(ProvenanceDataItem(
            self._add_name(names, "Times_the_dma_queue_was_overloaded"),
            dma_queue_overloaded,
            report=dma_queue_overloaded != 0,
            message=(
                "The DMA queue for {} on {}, {}, {} overloaded on {} "
                "occasions. This is often a sign that the system is running "
                "too quickly for the number of neurons per core.  Please "
                "increase the machine time step or time_scale_factor or "
                "decrease the number of neurons per core.".format(
                    label, x, y, p, dma_queue_overloaded))))

        data_items.append(ProvenanceDataItem(
            self._add_name(names, "Times_the_timer_tic_over_ran"),
            number_of_times_timer_tic_over_ran,
            report=number_of_times_timer_tic_over_ran > 4,
            message=(
                "A Timer tick callback was still executing when the next "
                "timer tick callback was fired off for {} on {}, {}, {}, {} "
                "times. This is a sign of the system being overloaded and "
                "therefore the results are likely incorrect.  Please increase "
                "the machine time step or time_scale_factor or decrease the "
                "number of neurons per core".format(
                    label, x, y, p, number_of_times_timer_tic_over_ran))))

        data_items.append(ProvenanceDataItem(
            self._add_name(names, "max_number_of_times_timer_tic_over_ran"),
            max_number_of_times_timer_tic_over_ran,
            report=max_number_of_times_timer_tic_over_ran > 4,
            message=(
                "The timer for {} on {}, {}, {} fell behind by up to {} "
                "ticks. This is a sign of the system being overloaded and "
                "therefore the results are likely incorrect. Please increase "
                "the machine time step or time_scale_factor or decrease the "
                "number of neurons per core".format(
                    label, x, y, p, max_number_of_times_timer_tic_over_ran))))

        return data_items

    def _get_remaining_provenance_data_items(self, provenance_data):
        return provenance_data[
            len(self.PROVENANCE_DATA_ENTRIES):]

    def get_provenance_data_from_machine(self, transceiver, placement):
        provenance_data = self._read_provenance_data(
            transceiver, placement)
        return self._read_basic_provenance_items(provenance_data, placement)



from six import add_metaclass

from spinn_utilities.abstract_base import AbstractBase, abstractmethod


@add_metaclass(AbstractBase)
class AbstractHasAssociatedBinary(object):

    __slots__ = ()

    @abstractmethod
    def get_binary_file_name(self):
        """ Get the binary name to be run for this vertex

        :rtype: str
        """

    @abstractmethod
    def get_binary_start_type(self):
        """ Get the start type of the binary to be run

        :rtype:\
            :py:class:`spinn_front_end_common.utilities.utility_objs.executable_start_type.ExecutableStartType`
        """



from pacman.executor.injection_decorator import inject_items
from pacman.model.decorators import overrides
from pacman.model.graphs.machine import MachineVertex
from pacman.model.resources import CPUCyclesPerTickResource, DTCMResource
from pacman.model.resources import IPtagResource, ResourceContainer
from pacman.model.resources import SDRAMResource

from spinn_front_end_common.interface.provenance \
    import ProvidesProvenanceDataFromMachineImpl
from spinn_front_end_common.interface.simulation.simulation_utilities \
    import get_simulation_header_array
from spinn_front_end_common.abstract_models \
    import AbstractGeneratesDataSpecification, AbstractHasAssociatedBinary, \
    AbstractSupportsDatabaseInjection
from spinn_front_end_common.utilities.utility_objs \
    import ProvenanceDataItem, ExecutableStartType
from spinn_front_end_common.utilities.constants \
    import SYSTEM_BYTES_REQUIREMENT

from spinnman.messages.eieio import EIEIOType

from enum import Enum
import struct


class LivePacketGatherMachineVertex(
        MachineVertex, ProvidesProvenanceDataFromMachineImpl,
        AbstractGeneratesDataSpecification, AbstractHasAssociatedBinary,
        AbstractSupportsDatabaseInjection):

    _LIVE_DATA_GATHER_REGIONS = Enum(
        value="LIVE_DATA_GATHER_REGIONS",
        names=[('SYSTEM', 0),
               ('CONFIG', 1),
               ('PROVENANCE', 2)])

    N_ADDITIONAL_PROVENANCE_ITEMS = 2
    _CONFIG_SIZE = 48
    _PROVENANCE_REGION_SIZE = 8

    def __init__(
            self, label, use_prefix=False, key_prefix=None, prefix_type=None,
            message_type=EIEIOType.KEY_32_BIT, right_shift=0,
            payload_as_time_stamps=True, use_payload_prefix=True,
            payload_prefix=None, payload_right_shift=0,
            number_of_packets_sent_per_time_step=0,
            hostname=None, port=None, strip_sdp=None, board_address=None,
            tag=None, constraints=None):
        # inheritance
        MachineVertex.__init__(self, label, constraints=constraints)

        self._resources_required = ResourceContainer(
            cpu_cycles=CPUCyclesPerTickResource(self.get_cpu_usage()),
            dtcm=DTCMResource(self.get_dtcm_usage()),
            sdram=SDRAMResource(self.get_sdram_usage()),
            iptags=[IPtagResource(
                ip_address=hostname, port=port,
                strip_sdp=strip_sdp, tag=tag,
                traffic_identifier="LPG_EVENT_STREAM")])

        # app specific data items
        self._use_prefix = use_prefix
        self._key_prefix = key_prefix
        self._prefix_type = prefix_type
        self._message_type = message_type
        self._right_shift = right_shift
        self._payload_as_time_stamps = payload_as_time_stamps
        self._use_payload_prefix = use_payload_prefix
        self._payload_prefix = payload_prefix
        self._payload_right_shift = payload_right_shift
        self._number_of_packets_sent_per_time_step = \
            number_of_packets_sent_per_time_step

    @property
    @overrides(ProvidesProvenanceDataFromMachineImpl._provenance_region_id)
    def _provenance_region_id(self):
        return self._LIVE_DATA_GATHER_REGIONS.PROVENANCE.value

    @property
    @overrides(ProvidesProvenanceDataFromMachineImpl._n_additional_data_items)
    def _n_additional_data_items(self):
        return self.N_ADDITIONAL_PROVENANCE_ITEMS

    @property
    @overrides(MachineVertex.resources_required)
    def resources_required(self):
        return self._resources_required

    @property
    @overrides(AbstractSupportsDatabaseInjection.is_in_injection_mode)
    def is_in_injection_mode(self):
        return True

    @overrides(ProvidesProvenanceDataFromMachineImpl.
               get_provenance_data_from_machine)
    def get_provenance_data_from_machine(self, transceiver, placement):
        """ Get provenance from the machine

        :param transceiver: spinnman interface to the machine
        :param placement: the location of this vertex on the machine
        """
        provenance_data = self._read_provenance_data(transceiver, placement)
        provenance_items = self._read_basic_provenance_items(
            provenance_data, placement)
        provenance_data = self._get_remaining_provenance_data_items(
            provenance_data)
        _, _, _, _, names = self._get_placement_details(placement)

        provenance_items.append(ProvenanceDataItem(
            self._add_name(names, "lost_packets_without_payload"),
            provenance_data[0],
            report=provenance_data[0] > 0,
            message=(
                "The live packet gatherer has lost {} packets which have "
                "payloads during its execution. Try increasing the machine "
                "time step or increasing the time scale factor. If you are "
                "running in real time, try reducing the number of vertices "
                "which are feeding this live packet gatherer".format(
                    provenance_data[0]))))
        provenance_items.append(ProvenanceDataItem(
            self._add_name(names, "lost_packets_with_payload"),
            provenance_data[1],
            report=provenance_data[1] > 0,
            message=(
                "The live packet gatherer has lost {} packets which do not "
                "have payloads during its execution. Try increasing the "
                "machine time step or increasing the time scale factor. If "
                "you are running in real time, try reducing the number of "
                "vertices which are feeding this live packet gatherer".format(
                    provenance_data[1]))))

        return provenance_items

    @overrides(AbstractHasAssociatedBinary.get_binary_file_name)
    def get_binary_file_name(self):
        return 'live_packet_gather.aplx'

    @overrides(AbstractHasAssociatedBinary.get_binary_start_type)
    def get_binary_start_type(self):
        return ExecutableStartType.USES_SIMULATION_INTERFACE

    @inject_items({
        "machine_time_step": "MachineTimeStep",
        "time_scale_factor": "TimeScaleFactor",
        "tags": "MemoryTags"})
    @overrides(
        AbstractGeneratesDataSpecification.generate_data_specification,
        additional_arguments={
            "machine_time_step", "time_scale_factor", "tags"
        })
    def generate_data_specification(
            self, spec, placement, machine_time_step, time_scale_factor,
            tags):

        spec.comment("\n*** Spec for LivePacketGather Instance ***\n\n")

        # Construct the data images needed for the Neuron:
        self._reserve_memory_regions(spec)
        self._write_setup_info(spec, machine_time_step, time_scale_factor)
        self._write_configuration_region(
            spec, tags.get_ip_tags_for_vertex(self))

        # End-of-Spec:
        spec.end_specification()

    def _reserve_memory_regions(self, spec):
        """ Reserve SDRAM space for memory areas
        """

        spec.comment("\nReserving memory space for data regions:\n\n")

        # Reserve memory:
        spec.reserve_memory_region(
            region=(
                LivePacketGatherMachineVertex.
                _LIVE_DATA_GATHER_REGIONS.SYSTEM.value),
            size=SYSTEM_BYTES_REQUIREMENT,
            label='system')
        spec.reserve_memory_region(
            region=(
                LivePacketGatherMachineVertex.
                _LIVE_DATA_GATHER_REGIONS.CONFIG.value),
            size=self._CONFIG_SIZE, label='config')
        self.reserve_provenance_data_region(spec)

    def _write_configuration_region(self, spec, iptags):
        """ writes the configuration region to the spec

        :param spec: the spec object for the dsg
        :type spec: \
                    :py:class:`spinn_storage_handlers.FileDataWriter`
        :param iptags: The set of ip tags assigned to the object
        :type iptags: iterable of :py:class:`spinn_machine.tags.IPTag`
        :raise DataSpecificationException: when something goes wrong with the\
                    dsg generation
        """
        spec.switch_write_focus(
            region=(
                LivePacketGatherMachineVertex.
                _LIVE_DATA_GATHER_REGIONS.CONFIG.value))

        # has prefix
        if self._use_prefix:
            spec.write_value(data=1)
        else:
            spec.write_value(data=0)

        # prefix
        if self._key_prefix is not None:
            spec.write_value(data=self._key_prefix)
        else:
            spec.write_value(data=0)

        # prefix type
        if self._prefix_type is not None:
            spec.write_value(data=self._prefix_type.value)
        else:
            spec.write_value(data=0)

        # packet type
        spec.write_value(data=self._message_type.value)

        # right shift
        spec.write_value(data=self._right_shift)

        # payload as time stamp
        if self._payload_as_time_stamps:
            spec.write_value(data=1)
        else:
            spec.write_value(data=0)

        # payload has prefix
        if self._use_payload_prefix:
            spec.write_value(data=1)
        else:
            spec.write_value(data=0)

        # payload prefix
        if self._payload_prefix is not None:
            spec.write_value(data=self._payload_prefix)
        else:
            spec.write_value(data=0)

        # right shift
        spec.write_value(data=self._payload_right_shift)

        # SDP tag
        iptag = iter(iptags).next()
        spec.write_value(data=iptag.tag)
        spec.write_value(struct.unpack("<H", struct.pack(
            "<BB", iptag.destination_y, iptag.destination_x))[0])

        # number of packets to send per time stamp
        spec.write_value(data=self._number_of_packets_sent_per_time_step)

    def _write_setup_info(self, spec, machine_time_step, time_scale_factor):
        """ Write basic info to the system region

        """

        # Write this to the system region (to be picked up by the simulation):
        spec.switch_write_focus(
            region=(LivePacketGatherMachineVertex.
                    _LIVE_DATA_GATHER_REGIONS.SYSTEM.value))
        spec.write_array(get_simulation_header_array(
            self.get_binary_file_name(), machine_time_step, time_scale_factor))

    @staticmethod
    def get_cpu_usage():
        """ Get the CPU used by this vertex

        :return:  0
        :rtype: int
        """
        return 0

    @staticmethod
    def get_sdram_usage():
        """ Get the SDRAM used by this vertex

        """
        return (
            SYSTEM_BYTES_REQUIREMENT +
            LivePacketGatherMachineVertex._CONFIG_SIZE +
            LivePacketGatherMachineVertex.get_provenance_data_size(
                LivePacketGatherMachineVertex.N_ADDITIONAL_PROVENANCE_ITEMS))

    @staticmethod
    def get_dtcm_usage():
        """ Get the DTCM used by this vertex

        """
        return LivePacketGatherMachineVertex._CONFIG_SIZE



# general imports
import os
import itertools
import json
import string


class ProvenanceJSONWriter(object):
    """ Write provenance data into JSON
    """

    __slots__ = []

    VALID_CHARS = frozenset(
        "-_.() {}{}".format(string.ascii_letters, string.digits))

    def __call__(self, provenance_data_items, provenance_data_path):

        # Group data by the first name
        items = sorted(provenance_data_items, key=lambda item: item.names[0])
        for name, group in itertools.groupby(
                items, lambda item: item.names[0]):

            filename = "".join(
                c if c in self.VALID_CHARS else '_' for c in name)

            # generate file path for xml
            file_path = os.path.join(
                provenance_data_path, "{}.json".format(filename))
            count = 2
            while os.path.exists(file_path):
                file_path = os.path.join(
                    provenance_data_path, "{}_{}.json".format(filename, count))
                count += 1

            # Create a root node
            root = dict()

            # Go through the items and add them
            for item in group:

                # Add the "categories" for the item (any name between the first
                # and last)
                super_element = root
                for cat_name in item.names[1:-1]:

                    if (cat_name in super_element and
                            isinstance(super_element[cat_name], dict)):

                        # If there is already a category of this name under the
                        # super element, use it
                        super_element = super_element[cat_name]
                    else:

                        # Otherwise, create a new category under the super
                        # element
                        super_element[cat_name] = dict()
                        super_element = super_element[cat_name]

                # Add the item
                super_element[item.names[-1]] = str(item.value)

            # write json form into file provided
            writer = open(file_path, "w")
            json.dump(root, writer, indent=4, separators=(',', ': '))
            writer.flush()
            writer.close()



# pacman imports
from pacman.model.constraints.placer_constraints\
    import RadialPlacementFromChipConstraint
from pacman.model.decorators import overrides
from pacman.model.graphs.application import ApplicationVertex
from pacman.model.resources import CPUCyclesPerTickResource, DTCMResource
from pacman.model.resources import IPtagResource, ResourceContainer
from pacman.model.resources import SDRAMResource

# spinn front end imports
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from .live_packet_gather_machine_vertex import LivePacketGatherMachineVertex
from spinn_front_end_common.abstract_models \
    import AbstractGeneratesDataSpecification, AbstractHasAssociatedBinary
from spinn_front_end_common.utilities.utility_objs import ExecutableStartType

# spinnman imports
from spinnman.messages.eieio import EIEIOType, EIEIOPrefix


class LivePacketGather(
        AbstractGeneratesDataSpecification, AbstractHasAssociatedBinary,
        ApplicationVertex):
    """ A model which stores all the events it receives during a timer tick\
        and then compresses them into Ethernet packets and sends them out of\
        a spinnaker machine.
    """

    def __init__(
            self, hostname, port, board_address=None, tag=None,
            strip_sdp=True, use_prefix=False, key_prefix=None,
            prefix_type=None, message_type=EIEIOType.KEY_32_BIT, right_shift=0,
            payload_as_time_stamps=True, use_payload_prefix=True,
            payload_prefix=None, payload_right_shift=0,
            number_of_packets_sent_per_time_step=0, constraints=None,
            label=None):
        """
        """
        if ((message_type == EIEIOType.KEY_PAYLOAD_32_BIT or
             message_type == EIEIOType.KEY_PAYLOAD_16_BIT) and
                use_payload_prefix and payload_as_time_stamps):
            raise ConfigurationException(
                "Timestamp can either be included as payload prefix or as "
                "payload to each key, not both")
        if ((message_type == EIEIOType.KEY_32_BIT or
             message_type == EIEIOType.KEY_16_BIT) and
                not use_payload_prefix and payload_as_time_stamps):
            raise ConfigurationException(
                "Timestamp can either be included as payload prefix or as"
                " payload to each key, but current configuration does not "
                "specify either of these")
        if (not isinstance(prefix_type, EIEIOPrefix) and
                prefix_type is not None):
            raise ConfigurationException(
                "the type of a prefix type should be of a EIEIOPrefix, "
                "which can be located in :"
                "SpinnMan.messages.eieio.eieio_prefix_type")

        if label is None:
            label = "Live Packet Gatherer"

        ApplicationVertex.__init__(self, label, constraints, 1)

        # Try to place this near the Ethernet
        self.add_constraint(RadialPlacementFromChipConstraint(0, 0))

        # storage objects
        self._iptags = None

        # tag info
        self._ip_address = hostname
        self._port = port
        self._board_address = board_address
        self._tag = tag
        self._strip_sdp = strip_sdp

        # eieio info
        self._prefix_type = prefix_type
        self._use_prefix = use_prefix
        self._key_prefix = key_prefix
        self._message_type = message_type
        self._right_shift = right_shift
        self._payload_as_time_stamps = payload_as_time_stamps
        self._use_payload_prefix = use_payload_prefix
        self._payload_prefix = payload_prefix
        self._payload_right_shift = payload_right_shift
        self._number_of_packets_sent_per_time_step = \
            number_of_packets_sent_per_time_step

    @overrides(ApplicationVertex.create_machine_vertex)
    def create_machine_vertex(
            self, vertex_slice, resources_required,
            label=None, constraints=None):
        return LivePacketGatherMachineVertex(
            label, self._use_prefix, self._key_prefix, self._prefix_type,
            self._message_type, self._right_shift,
            self._payload_as_time_stamps, self._use_payload_prefix,
            self._payload_prefix, self._payload_right_shift,
            self._number_of_packets_sent_per_time_step,
            hostname=self._ip_address, port=self._port,
            strip_sdp=self._strip_sdp, board_address=self._board_address,
            constraints=constraints)

    @overrides(AbstractHasAssociatedBinary.get_binary_file_name)
    def get_binary_file_name(self):
        return 'live_packet_gather.aplx'

    @overrides(AbstractHasAssociatedBinary.get_binary_start_type)
    def get_binary_start_type(self):
        return ExecutableStartType.USES_SIMULATION_INTERFACE

    @property
    @overrides(ApplicationVertex.n_atoms)
    def n_atoms(self):
        return 1

    @overrides(ApplicationVertex.get_resources_used_by_atoms)
    def get_resources_used_by_atoms(self, vertex_slice):
        return ResourceContainer(
            sdram=SDRAMResource(
                LivePacketGatherMachineVertex.get_sdram_usage()),
            dtcm=DTCMResource(LivePacketGatherMachineVertex.get_dtcm_usage()),
            cpu_cycles=CPUCyclesPerTickResource(
                LivePacketGatherMachineVertex.get_cpu_usage()),
            iptags=[IPtagResource(
                ip_address=self._ip_address, port=self._port,
                strip_sdp=self._strip_sdp, tag=self._tag,
                traffic_identifier="LPG_EVENT_STREAM")])

    @overrides(AbstractGeneratesDataSpecification.generate_data_specification)
    def generate_data_specification(self, spec, placement):

        # generate spec for the machine vertex
        placement.vertex.generate_data_specification(spec, placement)



from spinn_utilities.progress_bar import ProgressBar

# front end common imports
from spinn_front_end_common.utilities.utility_objs import ProvenanceDataItem

import logging

logger = logging.getLogger(__name__)


class RouterProvenanceGatherer(object):
    """
    RouterProvenanceGatherer: gathers diagnostics from the
    routers.
    """

    __slots__ = [
        # int for how many packets were sent
        '_total_sent_packets',

        # how many new packets were received
        '_total_new_packets',

        # how many dropped packets
        '_total_dropped_packets',

        # total missed dropped packets
        '_total_missed_dropped_packets',

        # total lost dropped packets
        '_total_lost_dropped_packets'

        # total
    ]

    def __call__(
            self, transceiver, machine, router_tables, has_ran,
            provenance_data_objects=None):
        """
        :param transceiver: the SpiNNMan interface object
        :param machine: the python representation of the spinnaker machine
        :param router_tables: the router tables that have been generated
        :param has_ran: token that states that the simulation has ran
        """

        if not has_ran:
            logger.warning("{} skipped as nothing has run "
                           "".format(self.__class__.__name__))
            return list()

        self._total_sent_packets = 0
        self._total_new_packets = 0
        self._total_dropped_packets = 0
        self._total_missed_dropped_packets = 0
        self._total_lost_dropped_packets = 0

        if provenance_data_objects is not None:
            prov_items = provenance_data_objects
        else:
            prov_items = list()

        prov_items.extend(self._write_router_provenance_data(
            router_tables, machine, transceiver))

        prov_items.append(ProvenanceDataItem(
            ["router_provenance", "total_multi_cast_sent_packets"],
            self._total_sent_packets))
        prov_items.append(ProvenanceDataItem(
            ["router_provenance", "total_created_packets"],
            self._total_new_packets))
        prov_items.append(ProvenanceDataItem(
            ["router_provenance", "total_dropped_packets"],
            self._total_dropped_packets))
        prov_items.append(ProvenanceDataItem(
            ["router_provenance", "total_missed_dropped_packets"],
            self._total_missed_dropped_packets))
        prov_items.append(ProvenanceDataItem(
            ["router_provenance", "total_lost_dropped_packets"],
            self._total_lost_dropped_packets))

        return prov_items

    def _write_router_provenance_data(self, router_tables, machine, txrx):
        """ Writes the provenance data of the router diagnostics

        :param router_tables: the routing tables generated by pacman
        :param machine: the spinnMachine object
        :param txrx: the transceiver object
        """
        progress = ProgressBar(machine.n_chips*2, "Getting Router Provenance")

        # acquire diagnostic data
        items = list()
        seen_chips = set()

        for router_table in sorted(
                router_tables.routing_tables,
                key=lambda table: (table.x, table.y)):
            self._write_router_table_diagnostic(
                txrx, machine, router_table.x, router_table.y, seen_chips,
                router_table, items)
            progress.update()

        for chip in sorted(machine.chips, key=lambda c: (c.x, c.y)):
            self._write_router_chip_diagnostic(txrx, chip, seen_chips, items)
            progress.update()
        progress.end()
        return items

    def _write_router_table_diagnostic(self, txrx, machine, x, y, seen_chips,
                                       router_table, items):
        if not machine.get_chip_at(x, y).virtual:
            try:
                router_diagnostic = txrx.get_router_diagnostics(x, y)
                seen_chips.add((x, y))
                reinjector_status = txrx.get_reinjection_status(x, y)
                items.extend(self._write_router_diagnostics(
                    x, y, router_diagnostic, reinjector_status, True,
                    router_table))
                self._add_totals(router_diagnostic, reinjector_status)
            except Exception as e:
                logger.warn(
                    "Could not read routing diagnostics from {}, {}: {}"
                    .format(x, y, e))

    def _write_router_chip_diagnostic(self, txrx, chip, seen_chips, items):
        if not chip.virtual and (chip.x, chip.y) not in seen_chips:
            try:
                diagnostic = txrx.get_router_diagnostics(chip.x, chip.y)

                if (diagnostic.n_dropped_multicast_packets or
                        diagnostic.n_local_multicast_packets or
                        diagnostic.n_external_multicast_packets):
                    reinjector_status = txrx.get_reinjection_status(
                            chip.x, chip.y)
                    items.extend(self._write_router_diagnostics(
                            chip.x, chip.y, diagnostic, reinjector_status,
                            False, None))
                    self._add_totals(diagnostic, reinjector_status)
            except Exception:
                # There could be issues with unused chips - don't worry!
                pass

    def _add_totals(self, router_diagnostic, reinjector_status):
        self._total_sent_packets += (
            router_diagnostic.n_local_multicast_packets +
            router_diagnostic.n_external_multicast_packets)
        self._total_new_packets += router_diagnostic.n_local_multicast_packets
        self._total_dropped_packets += (
            router_diagnostic.n_dropped_multicast_packets)
        if reinjector_status is not None:
            self._total_missed_dropped_packets += (
                reinjector_status.n_missed_dropped_packets)
            self._total_lost_dropped_packets += (
                reinjector_status.n_dropped_packet_overflows)
        else:
            self._total_lost_dropped_packets += (
                router_diagnostic.n_dropped_multicast_packets)

    @staticmethod
    def _add_name(names, name):
        new_names = list(names)
        new_names.append(name)
        return new_names

    def _write_router_diagnostics(
            self, x, y, router_diagnostic, reinjector_status, expected,
            router_table):
        """ Stores router diagnostics as a set of provenance data items

        :param x: x coord of the router in question
        :param y: y coord of the router in question
        :param router_diagnostic: the router diagnostic object
        :param reinjector_status: the data gained from the reinjector
        :param router_table: the router table generated by the PACMAN tools
        """
        names = list()
        names.append("router_provenance")
        if expected:
            names.append("expected_routers")
        else:
            names.append("unexpected_routers")
        names.append("router_at_chip_{}_{}".format(x, y))

        items = list()

        items.append(ProvenanceDataItem(
            self._add_name(names, "Local_Multicast_Packets"),
            str(router_diagnostic.n_local_multicast_packets)))
        items.append(ProvenanceDataItem(
            self._add_name(names, "External_Multicast_Packets"),
            str(router_diagnostic.n_external_multicast_packets)))
        items.append(ProvenanceDataItem(
            self._add_name(names, "Dropped_Multicast_Packets"),
            str(router_diagnostic.n_dropped_multicast_packets),
            report=(
                router_diagnostic.n_dropped_multicast_packets > 0 and
                reinjector_status is None),
            message=(
                "The router on {}, {} has dropped {} multicast route packets. "
                "Try increasing the machine_time_step and/or the time scale "
                "factor or reducing the number of atoms per core."
                .format(x, y, router_diagnostic.n_dropped_multicast_packets))))
        items.append(ProvenanceDataItem(
            self._add_name(
                names, "Dropped_Multicast_Packets_via_local_transmission"),
            str(router_diagnostic.user_3),
            report=(router_diagnostic.user_3 > 0),
            message=(
                "The router on {}, {} has dropped {} multicast packets that"
                " were transmitted by local cores. This occurs where the "
                "router has no entry associated with the multi-cast key. "
                "Try investigating the keys allocated to the vertices and "
                "the router table entries for this chip.".format(
                    x, y, router_diagnostic.user_3))))
        items.append(ProvenanceDataItem(
            self._add_name(names, "default_routed_external_multicast_packets"),
            str(router_diagnostic.user_2),
            report=(router_diagnostic.user_2 > 0 and
                    ((router_table is not None and
                      router_table.number_of_defaultable_entries == 0) or
                     router_table is None)),
            message=(
                "The router on {}, {} has default routed {} multicast packets,"
                " but the router table did not expect any default routed "
                "packets. This occurs where the router has no entry"
                " associated with the multi-cast key. "
                "Try investigating the keys allocated to the vertices and "
                "the router table entries for this chip.".format(
                    x, y, router_diagnostic.user_2))))

        items.append(ProvenanceDataItem(
            self._add_name(names, "Local_P2P_Packets"),
            str(router_diagnostic.n_local_peer_to_peer_packets)))
        items.append(ProvenanceDataItem(
            self._add_name(names, "External_P2P_Packets"),
            str(router_diagnostic.n_external_peer_to_peer_packets)))
        items.append(ProvenanceDataItem(
            self._add_name(names, "Dropped_P2P_Packets"),
            str(router_diagnostic.n_dropped_peer_to_peer_packets)))
        items.append(ProvenanceDataItem(
            self._add_name(names, "Local_NN_Packets"),
            str(router_diagnostic.n_local_nearest_neighbour_packets)))
        items.append(ProvenanceDataItem(
            self._add_name(names, "External_NN_Packets"),
            str(router_diagnostic.n_external_nearest_neighbour_packets)))
        items.append(ProvenanceDataItem(
            self._add_name(names, "Dropped_NN_Packets"),
            str(router_diagnostic.n_dropped_nearest_neighbour_packets)))
        items.append(ProvenanceDataItem(
            self._add_name(names, "Local_FR_Packets"),
            str(router_diagnostic.n_local_fixed_route_packets)))
        items.append(ProvenanceDataItem(
            self._add_name(names, "External_FR_Packets"),
            str(router_diagnostic.n_external_fixed_route_packets)))
        items.append(ProvenanceDataItem(
            self._add_name(names, "Dropped_FR_Packets"),
            str(router_diagnostic.n_dropped_fixed_route_packets)))
        if reinjector_status is not None:
            items.append(ProvenanceDataItem(
                self._add_name(names, "Received_For_Reinjection"),
                reinjector_status.n_dropped_packets))
            items.append(ProvenanceDataItem(
                self._add_name(names, "Missed_For_Reinjection"),
                reinjector_status.n_missed_dropped_packets,
                report=reinjector_status.n_missed_dropped_packets > 0,
                message=(
                    "The reinjector on {}, {} has missed {} packets.".format(
                        x, y, reinjector_status.n_missed_dropped_packets))))
            items.append(ProvenanceDataItem(
                self._add_name(names, "Reinjection_Overflows"),
                reinjector_status.n_dropped_packet_overflows,
                report=reinjector_status.n_dropped_packet_overflows > 0,
                message=(
                    "The reinjector on {}, {} has dropped {} packets.".format(
                        x, y, reinjector_status.n_dropped_packet_overflows))))
            items.append(ProvenanceDataItem(
                self._add_name(names, "Reinjected"),
                reinjector_status.n_reinjected_packets))
            items.append(ProvenanceDataItem(
                self._add_name(names, "Dumped_from_a_Link"),
                str(reinjector_status.n_link_dumps),
                report=reinjector_status.n_link_dumps > 0,
                message=(
                    "The reinjector on {}, {} has detected that {} packets "
                    "were dumped from a outgoing link of this chip's router."
                    " This often occurs when external devices are used in the "
                    "script but not connected to the communication fabric "
                    "correctly. These packets may have been reinjected "
                    "multiple times and so this number may be a overestimate."
                    .format(x, y, reinjector_status.n_link_dumps))))
            items.append(ProvenanceDataItem(
                self._add_name(names, "Dumped_from_a_processor"),
                str(reinjector_status.n_processor_dumps),
                report=reinjector_status.n_processor_dumps > 0,
                message=(
                    "The reinjector on {}, {} has detected that {} packets "
                    "were dumped from a core failing to take the packet."
                    " This often occurs when the executable has crashed or"
                    " has not been given a multicast packet callback. It can"
                    " also result from the core taking too long to process"
                    " each packet. These packets were reinjected and so this"
                    " number is likely a overestimate.".format(
                        x, y, reinjector_status.n_processor_dumps))))
        return items



from spinn_utilities.progress_bar import ProgressBar
from pacman.model.routing_tables import MulticastRoutingTable
from pacman.operations.algorithm_reports import reports

from spinn_front_end_common.utilities.exceptions import ConfigurationException

import logging
import os

logger = logging.getLogger(__name__)


class RoutingTableFromMachineReport(object):
    def __call__(
            self, report_default_directory, routing_tables, transceiver,
            app_id, has_loaded_routing_tables_flag):
        if not has_loaded_routing_tables_flag:
            raise ConfigurationException(
                "This report needs the routing tables to be loaded onto the "
                "machine before being executed.")

        tables = list(routing_tables.routing_tables)
        progress = ProgressBar(tables, "Reading Routing Tables from Machine")

        folder_name = os.path.join(
            report_default_directory, "routing_tables_from_machine")

        os.mkdir(folder_name)

        # generate a file for every multicast entry
        for routing_table in progress.over(tables):
            # get multicast entries from machine
            machine_routing_table = self._read_routing_table(
                transceiver, routing_table, app_id)
            reports._generate_routing_table(machine_routing_table, folder_name)

    def _read_routing_table(self, txrx, table, app_id):
        machine_routing_table = MulticastRoutingTable(table.x, table.y)
        for routing_entry in txrx.get_multicast_routes(
                table.x, table.y, app_id):
            machine_routing_table.add_multicast_routing_entry(routing_entry)
        return machine_routing_table



from six import add_metaclass

from spinn_utilities.abstract_base import AbstractBase, abstractmethod

from pacman.executor.injection_decorator \
    import supports_injection, inject


@supports_injection
@add_metaclass(AbstractBase)
class NeedsNMachineTimeSteps(object):
    """ A class that uses the number of machine time steps
    """

    __slots__ = ()

    @inject("TotalMachineTimeSteps")
    @abstractmethod
    def set_n_machine_time_steps(self, n_machine_time_steps):
        pass



from spinn_utilities.progress_bar import ProgressBar

# front end common imports
from spinn_front_end_common.utilities.exceptions import ConfigurationException

from spinnman.messages.scp.enums import Signal
from spinnman.model.enums import CPUState

# general imports
import logging

logger = logging.getLogger(__name__)


class LoadExecutableImages(object):
    __slots__ = []

    def __call__(self, executable_targets, app_id, transceiver,
                 loaded_application_data_token):
        """ Go through the executable targets and load each binary to \
            everywhere and then send a start request to the cores that \
            actually use it
        """

        if not loaded_application_data_token:
            raise ConfigurationException(
                "The token for having loaded the application data token is set"
                " to false and therefore I cannot run. Please fix and try "
                "again")

        progress = ProgressBar(
            executable_targets.total_processors + 1,
            "Loading executables onto the machine")

        for binary in executable_targets.binaries:
            progress.update(self._launch_binary(
                executable_targets, binary, transceiver, app_id))

        self._start_simulation(executable_targets, transceiver, app_id)
        progress.update()
        progress.end()

        return True

    def _launch_binary(self, executable_targets, binary, txrx, app_id):
        core_subset = executable_targets.get_cores_for_binary(binary)
        txrx.execute_flood(
            core_subset, binary, app_id, wait=True, is_filename=True)
        return len(core_subset)

    def _start_simulation(self, executable_targets, txrx, app_id):
        txrx.wait_for_cores_to_be_in_state(
            executable_targets.all_core_subsets, app_id, [CPUState.READY])
        txrx.send_signal(app_id, Signal.START)



from enum import Enum


class Simulator_State(Enum):
    """ Different States the SimulATOR COULD BE IN
    """
    INIT = (0, "init called")
    IN_RUN = (1, "inside run method")
    RUN_FOREVER = (2, "finish run method but running forever")
    FINISHED = (3, "run ended shutdown not called")
    SHUTDOWN = (4, "shutdown called")

    def __new__(cls, value, doc=""):
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, value, doc=""):
        self._value_ = value
        self.__doc__ = doc



from .database_connection import DatabaseConnection
from .database_reader import DatabaseReader
from .database_writer import DatabaseWriter

__all__ = ["DatabaseConnection", "DatabaseReader", "DatabaseWriter"]



from .channel_buffer_state import ChannelBufferState


class EndBufferingState(object):
    """ Stores the buffering state at the end of a simulation
    """

    __slots__ = [
        #  a list of channel state, where each channel is stored in a
        # ChannelBufferState object
        "_list_channel_buffer_state",

        # the final state stuff
        "_buffering_out_fsm_state"
    ]

    def __init__(
            self, buffering_out_fsm_state, list_channel_buffer_state):
        """

        :param buffering_out_fsm_state: Final sequence number received
        :param list_channel_buffer_state: a list of channel state, where each\
                channel is stored in a ChannelBufferState object
        """
        self._buffering_out_fsm_state = buffering_out_fsm_state
        self._list_channel_buffer_state = list_channel_buffer_state

    @property
    def buffering_out_fsm_state(self):
        return self._buffering_out_fsm_state

    def channel_buffer_state(self, i):
        return self._list_channel_buffer_state[i]

    def get_state_for_region(self, region_id):
        for state in self._list_channel_buffer_state:
            if state.region_id == region_id:
                return state
        return None

    def get_missing_info_for_region(self, region_id):
        state = self.get_state_for_region(region_id)
        return None if state is None else state.missing_info

    @staticmethod
    def size_of_region(n_regions_to_record):
        size_of_header = 8 + 4 * n_regions_to_record

        # add size needed for the data region addresses
        size_of_header += 4 * n_regions_to_record
        size_of_channel_state = ChannelBufferState.size_of_channel_state()
        return size_of_header + n_regions_to_record * size_of_channel_state



import numpy
import logging

import scipy.stats

logger = logging.getLogger(__name__)

# Define profiler time scale in ms
_MS_SCALE = (1.0 / 200000.0)

# Define the positions in tags
_START_TIME = 0
_DURATION = 1


class ProfileData(object):
    """ A container for profile data
    """

    START_TIME = 0
    DURATION = 1

    __slots__ = (

        # A dictionary of tag label to numpy array of start times and durations
        "_tags",

        # A list of tag labels indexed by the tag id
        "_tag_labels",

        # The maximum time recorded
        "_max_time"
    )

    def __init__(self, tag_labels):
        """

        :param tag_labels: A list of labels indexed by tag id
        :type tag_labels: list(str)
        """
        self._tag_labels = tag_labels
        self._tags = dict()

    def add_data(self, data):
        """ Add profiling data read from the profile section

        :param data: Data read from the profile section on the machine
        :type data: bytearray
        """
        samples = numpy.asarray(data, dtype="uint8").view(dtype="<u4")

        # Slice data to separate times, tags and flags
        sample_times = samples[::2]
        sample_tags_and_flags = samples[1::2]

        # Further split the tags and flags word into separate arrays of tags
        # and flags
        sample_tags = numpy.bitwise_and(sample_tags_and_flags, 0x7FFFFFFF)
        sample_flags = numpy.right_shift(sample_tags_and_flags, 31)

        # Find indices of samples relating to entries and exits
        sample_entry_indices = numpy.where(sample_flags == 1)
        sample_exit_indices = numpy.where(sample_flags == 0)

        # Convert count-down times to count up times from 1st sample
        sample_times = numpy.subtract(sample_times[0], sample_times)
        sample_times_ms = numpy.multiply(
            sample_times, _MS_SCALE, dtype=numpy.float)

        # Slice tags and times into entry and exits
        entry_tags = sample_tags[sample_entry_indices]
        entry_times_ms = sample_times_ms[sample_entry_indices]
        exit_tags = sample_tags[sample_exit_indices]
        exit_times_ms = sample_times_ms[sample_exit_indices]

        # Loop through unique tags
        unique_tags = numpy.unique(sample_tags)
        for tag in unique_tags:

            tag_label = self._tag_labels.get(tag, None)
            if tag_label is None:
                logger.warn("Unknown tag {} in profile data".format(tag))
                tag_label = "UNKNOWN"

            # Get indices where these tags occur
            tag_entry_indices = numpy.where(entry_tags == tag)
            tag_exit_indices = numpy.where(exit_tags == tag)

            # Use these to get subset for this tag
            tag_entry_times_ms = entry_times_ms[tag_entry_indices]
            tag_exit_times_ms = exit_times_ms[tag_exit_indices]

            # If the first exit is before the first
            # Entry, add a dummy entry at beginning
            if tag_exit_times_ms[0] < tag_entry_times_ms[0]:
                logger.warn("Profile starts mid-tag")
                tag_entry_times_ms = numpy.append(0.0, tag_entry_times_ms)

            if len(tag_entry_times_ms) > len(tag_exit_times_ms):
                logger.warn("profile finishes mid-tag")
                tag_entry_times_ms = tag_entry_times_ms[
                    :len(tag_exit_times_ms) - len(tag_entry_times_ms)]

            # Subtract entry times from exit times to get durations of each
            # call
            tag_durations_ms = numpy.subtract(
                tag_exit_times_ms, tag_entry_times_ms)

            # Add entry times and durations to dictionary
            self._tags[tag_label] = (tag_entry_times_ms, tag_durations_ms)

            # Keep track of the maximum time
            self._max_time = numpy.max(tag_entry_times_ms + tag_durations_ms)

    @property
    def tags(self):
        """ The tags recorded as labels

        :rtype: list of str
        """
        return self._tags.keys()

    def get_mean_ms(self, tag):
        """ Get the mean time in milliseconds spent on operations with the\
            given tag

        :param tag: The tag to get the mean time for
        :type tag: str
        :rtype: float
        """
        return numpy.average(self._tags[tag][_DURATION])

    def get_n_calls(self, tag):
        """ Get the number of times the given tag was recorded

        :param tag: The tag to get the number of calls of
        :type tag: str
        :rtype: int
        """
        return self._tags[tag][_DURATION].size

    def get_mean_n_calls_per_ts(self, tag, run_time_ms, machine_time_step_ms):
        """ Get the mean number of times the given tag was recorded per\
            timestep

        :param tag: The tag to get the data for
        :type tag: str
        :param machine_time_step_ms:\
            The time step of the simulation in microseconds
        :type machine_time_step_ms: int
        :param run_time_ms: The run time of the simulation in milliseconds
        :type run_time_ms: float
        :rtype: float
        """
        bins = numpy.arange(
            0, self._max_time + machine_time_step_ms, machine_time_step_ms)
        return numpy.average(numpy.histogram(
            self._tags[tag][_START_TIME], bins)[0])

    def get_mean_ms_per_ts(self, tag, run_time_ms, machine_time_step_ms):
        """ Get the mean time in milliseconds spent on operations with the\
            given tag per timestep

        :param tag: The tag to get the data for
        :type tag: str
        :param machine_time_step_ms:\
            The time step of the simulation in microseconds
        :type machine_time_step_ms: int
        :param run_time_ms: The run time of the simulation in milliseconds
        :type run_time_ms: float
        :rtype: float
        """
        bins = numpy.arange(
            0, self._max_time + machine_time_step_ms, machine_time_step_ms)
        mean_per_ts = scipy.stats.binned_statistic(
            self._tags[tag][_START_TIME], self._tags[tag][_DURATION],
            "mean", bins).statistic
        mean_per_ts[numpy.isnan(mean_per_ts)] = 0
        return numpy.average(
            mean_per_ts[numpy.logical_not(numpy.isnan(mean_per_ts))])



from spinn_front_end_common.utilities.notification_protocol \
    import NotificationProtocol as Notification

import logging


logger = logging.getLogger(__name__)


class NotificationProtocol(object):
    """ The notification protocol for external device interaction
    """

    __slots__ = [
        # the notification protocol for talking to external devices
        "_notification_protocol"
    ]

    def __call__(
            self, wait_for_read_confirmation,
            socket_addresses, database_file_path):

        # notification protocol
        self._notification_protocol = Notification(
            socket_addresses, wait_for_read_confirmation)
        self.send_read_notification(database_file_path)

        return self

    def wait_for_confirmation(self):
        """ Waits for devices to confirm they have read the database via the\
            notification protocol

        :rtype: None:
        """
        self._notification_protocol.wait_for_confirmation()

    def send_read_notification(self, database_directory):
        """ Send the read notifications via the notification protocol

        :param database_directory: the path to the database
        :rtype: None:
        """
        self._notification_protocol.send_read_notification(database_directory)

    def send_start_resume_notification(self):
        """ Send the start notifications via the notification protocol

        :rtype: None:
        """
        self._notification_protocol.send_start_resume_notification()

    def send_stop_pause_notification(self):
        """ Send the stop or pause notifications via the notification protocol

        :rtype: None:
        """
        self._notification_protocol.send_stop_pause_notification()

    def stop(self):
        """ Ends the notification protocol

        :rtype: None:
        """
        logger.debug("[data_base_thread] Stopping")
        self._notification_protocol.close()



from spinn_utilities.progress_bar import ProgressBar

# front end common imports
from spinn_front_end_common.interface.profiling.abstract_has_profile_data \
    import AbstractHasProfileData

import os
import logging

logger = logging.getLogger(__name__)


class ProfileDataGatherer(object):
    __slots__ = []

    def __call__(
            self, transceiver, placements, has_ran, provenance_file_path,
            run_time_ms, machine_time_step):
        """
        :param transceiver: the SpiNNMan interface object
        :param placements: The placements of the vertices
        :param has_ran: token that states that the simulation has ran
        :param provenance_file_path: The location to store the profile data
        :param run_time_ms: runtime in ms
        :param machine_time_step: machine time step in ms
        """

        machine_time_step_ms = machine_time_step / 1000

        if not has_ran:
            logger.warning("{} skipped as nothing has run "
                           "".format(self.__class__.__name__))
            return

        progress = ProgressBar(
            placements.n_placements, "Getting profile data")

        # retrieve provenance data from any cores that provide data
        for placement in progress.over(placements.placements):
            if isinstance(placement.vertex, AbstractHasProfileData):

                # get data
                profile_data = placement.vertex.get_profile_data(
                    transceiver, placement)

                if len(profile_data.tags) > 0:

                    max_tag_len = max([len(tag) for tag in profile_data.tags])

                    # write data
                    file_name = os.path.join(
                        provenance_file_path, "{}_{}_{}_profile.txt".format(
                            placement.x, placement.y, placement.p))

                    # set mode of the file based off if the file already exists
                    mode = "w"
                    if os.path.exists(file_name):
                        mode = "a"

                    # write profile data to file
                    with open(file_name, mode) as writer:
                        writer.write(
                            "{: <{}s} {: <7s} {: <14s} {: <14s} {: <14s}\n"
                            .format(
                                "tag", max_tag_len, "n_calls", "mean_ms",
                                "n_calls_per_ts", "mean_ms_per_ts"))
                        writer.write(
                            "{:-<{}s} {:-<7s} {:-<14s} {:-<14s} {:-<14s}\n"
                            .format(
                                "", max_tag_len, "", "", "", ""))
                        for tag in profile_data.tags:
                            writer.write(
                                "{: <{}s} {: >7d} {: >14.6f} {: >14.6f} "
                                "{: >14.6f}\n"
                                .format(
                                    tag, max_tag_len,
                                    profile_data.get_n_calls(tag),
                                    profile_data.get_mean_ms(tag),
                                    profile_data.get_mean_n_calls_per_ts(
                                        tag, run_time_ms,
                                        machine_time_step_ms),
                                    profile_data.get_mean_ms_per_ts(
                                        tag, run_time_ms,
                                        machine_time_step_ms)))



# spinn front end common imports
from spinn_front_end_common.utility_models.\
    chip_power_monitor_application_vertex import \
    ChipPowerMonitorApplicationVertex
from spinn_front_end_common.utility_models.\
    chip_power_monitor_machine_vertex import ChipPowerMonitorMachineVertex

# pacman imports
from pacman.model.graphs.common.slice import Slice
from pacman.model.constraints.placer_constraints.\
    chip_and_core_constraint import ChipAndCoreConstraint

# utils imports
from spinn_utilities.progress_bar import ProgressBar


class InsertChipPowerMonitorsToGraphs(object):
    """ function to add chip power monitors into a given graph
    """

    def __call__(
            self, machine, machine_graph, n_samples_per_recording,
            sampling_frequency, application_graph=None, graph_mapper=None):
        """ call that adds LPG vertices on Ethernet connected chips as\
            required.

        :param machine: the spinnaker machine as discovered
        :param application_graph: the application graph
        :param machine_graph: the machine graph
        :return: mapping between LPG params and LPG vertex
        """

        # create progress bar
        progress_bar = ProgressBar(
            len(list(machine.chips)),
            string_describing_what_being_progressed=(
                "Adding Chip power monitors to Graph"))

        for chip in progress_bar.over(machine.chips):

            # build constraint
            constraint = ChipAndCoreConstraint(chip.x, chip.y)

            # build machine vert
            machine_vertex = ChipPowerMonitorMachineVertex(
                label="chip_power_monitor_machine_vertex_for_chip({}:{})".
                format(chip.x, chip.y),
                sampling_frequency=sampling_frequency,
                n_samples_per_recording=n_samples_per_recording,
                constraints=[constraint])

            # add vert to graph
            machine_graph.add_vertex(machine_vertex)

            # deal with app graphs if needed
            if application_graph is not None:

                # build app vertex
                vertex_slice = Slice(0, 0)
                application_vertex = \
                    ChipPowerMonitorApplicationVertex(
                        label="chip_power_monitor_application_vertex_for"
                              "_chip({}:{})".format(chip.x, chip.y),
                        constraints=[constraint],
                        sampling_frequency=sampling_frequency,
                        n_samples_per_recording=n_samples_per_recording)

                # add to graph
                application_graph.add_vertex(application_vertex)

                # update graph mapper
                graph_mapper.add_vertex_mapping(
                    machine_vertex, vertex_slice, application_vertex)
